from fs_uae_launcher.ui.settings.OptionUI import OptionUI
from fsgs.FSGSDirectories import FSGSDirectories
import fsui as fsui
from fsui.extra.iconheader import NewIconHeader
from ...I18N import gettext


class WHDLoadSettingsPage(fsui.Panel):

    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        self.layout = fsui.VerticalLayout()
        # self.layout.set_padding(20, 20, 20, 20)

        self.icon_header = NewIconHeader(
            self, fsui.Icon("settings", "pkg:fs_uae_workspace"),
            gettext("WHDLoad Settings"),
            gettext("Options for WHDLoad support in FS-UAE Launcher"))
        self.layout.add(self.icon_header, fill=True, margin_bottom=20)

        def add_option(name):
            self.layout.add(OptionUI.create_group(self, name), fill=True,
                            margin_top=10, margin_bottom=10)

        label = fsui.MultiLineLabel(self, gettext(
            "The following options only apply when you use the automatic "
            "WHDLoad support in FS-UAE Launcher, for example in relation with "
            "the online game database."), 640)
        self.layout.add(label, fill=True, margin_top=0)

        # label = fsui.MultiLineLabel(self, gettext(
        #     " When you use your own hard drive installation with WHDLoad "
        #     "manually installed, you need to configure WHDLoad within the "
        #     "Amiga environment instead."), 640)
        # self.layout.add(label, fill=True, margin_top=10)

        add_option("whdload_splash_delay")

        label = fsui.Label(
            self, gettext("Directory for WHDLoad.key file (if you have it):"))
        self.layout.add(label, margin_top=10)

        hor_layout = fsui.HorizontalLayout()
        self.layout.add(hor_layout, margin_top=4, fill=True)
        label = fsui.Label(
            self, FSGSDirectories.get_base_dir())
        hor_layout.add_spacer(0, expand=True)
        hor_layout.add(label)



import fsui as fsui
from ...Config import Config
from ...Settings import Settings
from ...GamePaths import GamePaths
from .BottomPanel import BottomPanel
from ..Constants import Constants
from ..ImageLoader import ImageLoader
from .RatingButton import RatingButton
from .EditButton import EditButton
from .WebButton import WebButton
from ..Skin import Skin

BORDER = 20


class GameInfoPanel(BottomPanel):

    def __init__(self, parent, cover_on_right=True):
        BottomPanel.__init__(self, parent)
        Skin.set_background_color(self)
        # self.set_background_color((0xdd, 0xdd, 0xdd))

        self.cover_on_right = cover_on_right
        self.requests = None
        self.image = None

        self.layout = fsui.HorizontalLayout()

        # def get_min_height():
        #     return Constants.COVER_SIZE[1] + 2 * BORDER
        # self.layout.get_min_height = get_min_height

        self.layout.padding_left = BORDER // 2
        self.layout.padding_right = BORDER // 2
        self.layout.padding_top = BORDER + 2
        self.layout.padding_bottom = Skin.get_bottom_margin()

        self.default_image = fsui.Image("fs_uae_launcher:res/cover.png")
        # self.default_image.resize(Constants.COVER_SIZE)
        self.cover_overlay = fsui.Image(
            "fs_uae_launcher:res/cover_overlay.png")
        self.cover_overlay_square = fsui.Image(
            "fs_uae_launcher:res/cover_overlay_square.png")

        if not self.cover_on_right:
            self.layout.add_spacer(Constants.COVER_SIZE[0])

        # self.panel = fsui.Panel(self)
        # self.panel.set_background_color((0xdd, 0xdd, 0xdd))
        # self.layout.add(self.panel, expand=True, fill=True)

        vert_layout = fsui.VerticalLayout()
        self.layout.add(vert_layout, expand=True, fill=True)
        # self.panel.layout.padding_top = 10
        if self.cover_on_right:
            vert_layout.padding_right = BORDER
        else:
            vert_layout.padding_left = BORDER

        vert_layout.add_spacer(58 + 3)
        hori_layout = fsui.HorizontalLayout()
        vert_layout.add(hori_layout, fill=True)

        hori_layout.add_spacer(0, expand=True)

        self.web_button = WebButton(self)
        hori_layout.add(self.web_button, margin_right=3)

        self.edit_button = EditButton(self)
        hori_layout.add(self.edit_button, margin_right=3)

        # def label_min_width():
        #     return 330
        # self.title_label = fsui.HeadingLabel(self)
        # self.title_label.get_min_width = label_min_width
        # vert_layout.add(self.title_label)

        # self.sub_title_label = fsui.Label(self)
        # self.sub_title_label.get_min_width = label_min_width
        # vert_layout.add(self.sub_title_label)

        self.title = ""
        self.sub_title = ""
        self.year = ""
        self.publisher = ""
        self.developer = ""
        self.companies = ""

        self.image_path = ""

        vert_layout.add_spacer(0, expand=True)

        hori_layout = fsui.HorizontalLayout()
        vert_layout.add(hori_layout, fill=True)

        # for rating in [1, 4, 5]:
        #     button = RatingButton(self, rating)
        #     hori_layout.add(button, margin_right=5, fill=True)

        hori_layout.add_spacer(0, expand=True)
        # self.launch_group = LaunchGroup(self)
        # hori_layout.add(self.launch_group, fill=True)

        if self.cover_on_right:
            self.layout.add_spacer(Constants.COVER_SIZE[0])

        self.load_info()
        Settings.add_listener(self)
        Config.add_listener(self)

        for key in ["database_url", "hol_url", "lemon_url", "mobygames_url",
                    "wikipedia_url", "year", "publisher", "developer"]:
            self.on_config(key, Config.get(key))

    def on_destroy(self):
        Config.remove_listener(self)
        Settings.remove_listener(self)

    def load_info(self):
        handler = GamePaths.current()
        name = handler.get_name()
        variant = handler.get_variant()
        if variant.startswith("Amiga, "):
            variant = variant[7:]
        self.title = name
        self.sub_title = variant
        self.sub_title = self.sub_title.replace(", ", " \u00b7 ")

        path = handler.get_cover_path()
        if path != self.image_path:
            self.image_path = path
            loader = ImageLoader.get()

            def on_load(request):
                if request.path != self.image_path:
                    return
                if request.image:
                    self.image = request.image
                else:
                    self.image = self.default_image
                self.image_path = request.path
                self.refresh()

            self.requests = loader.load_image(
                path, size=Constants.COVER_SIZE, on_load=on_load,
                is_cover=True)
            self.image = self.default_image

        self.refresh()
        self.layout.update()

    def on_setting(self, key, _):
        if key == "parent_uuid":
            self.load_info()
        if key == "config_name":
            self.load_info()

    def on_config(self, key, value):
        if key == "publisher":
            self.publisher = value
            self.update_companies()
        elif key == "developer":
            self.developer = value
            self.update_companies()
        elif key == "year":
            self.year = value

    def update_companies(self):
        companies = [x.strip() for x in self.publisher.split("/") if x.strip()]
        for developer in self.developer.split("/"):
            developer = developer.strip()
            if developer and developer not in companies:
                companies.append(developer)
        self.companies = " \u00b7 ".join(companies)

    def on_paint(self):
        dc = self.create_dc()
        self.draw_background(dc)

        if self.cover_on_right:
            x = self.get_size()[0] - 10 - Constants.COVER_SIZE[0]
            text_x = 10
        else:
            x = 10
            text_x = 10 + Constants.COVER_SIZE[0] + 20

        y = 2 + 20
        size = self.size

        image = self.image
        # dc.draw_image(image, x, y)
        if image.size[0] == image.size[1]:
            # cover_overlay = self.cover_overlay_square
            y_offset = 14
            # title_x = 10
        else:
            y_offset = 0
        cover_overlay = self.cover_overlay

        if image.size[0] == image.size[1]:
            dc.draw_rectangle(
                x + 1, y + 1, Constants.COVER_SIZE[0],
                Constants.COVER_SIZE[1], fsui.Color(0x0, 0x0, 0x0))
        dc.draw_image(image, x + 1, y + 1 + y_offset)
        dc.draw_image(cover_overlay, x - 10, y - 10)

        font = dc.get_font()
        font.set_bold(True)
        dc.set_font(font)

        dc.draw_text(self.title, text_x, y)
        # tw, th = \
        dc.measure_text(self.title)
        # y += int(th * 1.2)
        y += 24

        color = dc.get_text_color()
        color.mix(self.get_background_color(), 0.33)
        dc.set_text_color(color)

        if self.year:
            twy, thy = dc.measure_text(self.year)
            dc.draw_text(self.year, text_x, y)
            twy += 10
        else:
            twy = 0
        font.set_bold(False)
        dc.set_font(font)
        if self.companies:
            dc.draw_text(self.companies, text_x + twy, y)

        background = fsui.Color(0x00, 0x00, 0x00, 0x20)
        y = 80
        h = 30

        tw, th = dc.measure_text(self.sub_title)
        if self.cover_on_right:
            dc.draw_rectangle(
                text_x, y, size[0] - Constants.COVER_SIZE[0] - 20 - 20, h,
                background)
            dc.draw_text(self.sub_title, text_x + 6, y + (h - th) // 2)
        else:
            dc.draw_rectangle(
                text_x - 18, y, size[0] - text_x + 18, h, background)
            dc.draw_text(self.sub_title, text_x, y + (h - th) // 2)



import fsui as fsui
from .RemovableMediaGroup import RemovableMediaGroup
from .config.InputGroup import InputGroup
from .config.modelgroup import ModelGroup
from .ConfigGroup import ConfigGroup
from .Skin import Skin


class MainPanel(fsui.Panel):

    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        Skin.set_background_color(self)
        self.layout = fsui.VerticalLayout()

        # self.config_group = ConfigGroup(self)

        self.model_group = ModelGroup(self)
        self.removable_media_group = RemovableMediaGroup(self, 2)
        self.input_group = InputGroup(self)

        # self.layout.add(self.config_group, fill=True)
        # self.layout.add_spacer(Skin.EXTRA_GROUP_MARGIN)

        self.layout.add(self.model_group, fill=True)
        self.layout.add_spacer(Skin.EXTRA_GROUP_MARGIN)

        self.layout.add(self.removable_media_group, fill=True)
        self.layout.add_spacer(Skin.EXTRA_GROUP_MARGIN)

        self.layout.add(self.input_group, fill=True)



import json
from gzip import GzipFile
from io import StringIO
from urllib.parse import urlencode
from urllib.request import HTTPBasicAuthHandler, build_opener, Request
from fsgs.ogd.client import OGDClient


class OAGDClient(object):

    def __init__(self):
        self._server = None
        self._opener = None
        self._json = None
        self.data = b""

        # FIXME: we don't want dependency on Settings here
        from fs_uae_launcher.Settings import Settings
        self.username = "auth_token"
        self.password = Settings.get("database_auth")

    def server(self):
        return OGDClient.get_server()

    def opener(self):
        if self._opener:
            return self._opener
        server = self.server()
        auth_handler = HTTPBasicAuthHandler()
        auth_handler.add_password(
            realm="Open Amiga Game Database",
            uri="http://{0}".format(server), user=self.username,
            passwd=self.password)
        self._opener = build_opener(auth_handler)
        return self._opener

    def build_url(self, path, **kwargs):
        url = "http://{0}{1}".format(self.server(), path)
        if kwargs:
            url += "?" + urlencode(kwargs)
        return url

    def handle_response(self, response):
        self._json = None
        self.data = response.read()

        # print(dir(response.headers))
        try:
            getheader = response.headers.getheader
        except AttributeError:
            getheader = response.getheader
        content_encoding = getheader("content-encoding", "").lower()
        if content_encoding == "gzip":
            # data = zlib.decompress(data)
            fake_stream = StringIO(self.data)
            self.data = GzipFile(fileobj=fake_stream).read()

    def json_response(self):
        if self._json is None:
            self._json = json.loads(self.data.decode("UTF-8"))
        return self._json

    def get(self, key, default):
        doc = self.json_response()
        return doc.get(key, default)

    def get_request(self, url):
        opener = self.opener()
        request = Request(url)
        print("get_request:", url)
        request.add_header("Accept-Encoding", "gzip")
        response = opener.open(request)
        return self.handle_response(response)

    def rate_variant(self, variant_uuid, like=None, work=None):
        params = {
            "game": variant_uuid,
        }
        if like is not None:
            params["like"] = like
        if work is not None:
            params["work"] = work
        url = self.build_url("/api/1/rate_game", **params)
        self.get_request(url)



from fs_uae_launcher.Config import Config
from fs_uae_launcher.I18N import gettext
from fs_uae_launcher.Settings import Settings
from fs_uae_launcher.ui.IconButton import IconButton


class NewButton(IconButton):

    def __init__(self, parent):
        super().__init__(parent, "new_button.png")
        self.set_tooltip(gettext("New Configuration"))

    def on_activate(self):
        self.new_config()

    @staticmethod
    def new_config():
        Config.load_default_config()
        # Settings.set("config_changed", "1")
        Settings.set("parent_uuid", "")



# FSGS - Common functionality for Fengestad Game System.
# Copyright (C) 2013  Frode Solheim <frode-code@fengestad.no>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
Game Runner for Atari ST.

TODO:
- support saving disk changes
- support (persistent) save states
- check and/or fix support for vsync
- NTSC vs PAL?
- joystick support and keyboard joystick emulation
- mouse in port 0
- multiple floppies and floppy swapping

"""
import os
import io
import traceback
from fsgs.runner import GameRunner


ATARI_WIDTH = 832
# ATARI_HEIGHT = 576
ATARI_HEIGHT = 552

# TOS v1.02 (1987)(Atari Corp)(Mega ST)(GB)[MEGA TOS].img
# tos102uk.bin
TOS_102_UK = "87900a40a890fdf03bd08be6c60cc645855cbce5"

# TOS v1.04 (1989)(Atari Corp)(Mega ST)(GB)[Rainbow TOS].img
# tos104uk.bin
TOS_104_UK = "9526ef63b9cb1d2a7109e278547ae78a5c1db6c6"


# noinspection PyAttributeOutsideInit
class AtariSTRunner(GameRunner):

    def prepare(self):
        self.temp_dir = self.create_temp_dir("hatari")
        config_file = os.path.join(self.temp_dir.path, "hatari.cfg")
        with io.open(config_file, "w", encoding="UTF-8") as f:
            self.configure(f)
        self.add_arg("--configfile", config_file)
        self.add_arg("--statusbar", "0")
        self.add_arg("--drive-led", "0")

    def configure(self, f):
        # floppies_dir = os.path.join(self.context.temp.dir('uae'), 'Disks')
        floppies_dir = os.path.join(self.temp_dir.path, "Disks")
        os.makedirs(floppies_dir)
        original_floppies = self.prepare_floppies()
        floppies = []
        for p in original_floppies:
            dest_path = os.path.join(floppies_dir, os.path.basename(p))
            self.fsgs.file.copy_game_file(p, dest_path)
            floppies.append(dest_path)

        # self.changes = ChangeHandler(floppies_dir)
        # self.changes.init(os.path.join(self.context.game.state_dir, 'Disks'))

        num_floppy_drives = 2
        # sorted_floppies = self.sort_floppies(floppies)
        sorted_floppies = floppies
        inserted_floppies = sorted_floppies[:num_floppy_drives]

        f.write("[Floppy]\n")
        f.write("szDiskAFileName = {path}\n".format(
                path=inserted_floppies[0]))
        if len(inserted_floppies) > 1:
            f.write("szDiskBFileName = {path}\n".format(
                    path=inserted_floppies[1]))

        bios_path = self.prepare_tos()
        f.write("[ROM]\n")
        f.write("szTosImageFileName = {path}\n".format(path=bios_path))

        f.write("[System]\n")
        # don't patch Timer-D to reduce number of interrupts
        f.write("bPatchTimerD = FALSE\n")
        # f.write("bFastBoot = FALSE\n")

        f.write("[Sound]\n")
        if self.use_audio_frequency():
            f.write("nPlaybackFreq = {0}\n".format(self.use_audio_frequency()))

        if self.use_fullscreen():
            self.configure_fullscreen_scaling()

    def prepare_floppies(self):
        floppies = []
        # media_dir = os.path.dirname(self.context.game.file)
        # base_match = self.extract_match_name(os.path.basename(
        #         self.context.game.file))
        # for name in os.listdir(media_dir):
        #     dummy, ext = os.path.splitext(name)
        #     if ext.lower() not in ['.st']:
        #         continue
        #     match = self.extract_match_name(name)
        #     if base_match == match:
        #         floppies.append(os.path.join(media_dir, name))
        #         #floppies.append(name)
        if self.config["floppy_drive_0"]:
            floppies.append(self.config["floppy_drive_0"])
        return floppies

    def configure_fullscreen_scaling(self):
        sx, sy, sw, sh = 0, 0, ATARI_WIDTH, ATARI_HEIGHT
        try:
            viewport = self.config["viewport"]
            if viewport:
                sx, sy, sw, sh = viewport.rsplit("=", 1)[-1].strip().split(" ")
                sx, sy, sw, sh = int(sx), int(sy), int(sw), int(sh)
        except Exception:
            traceback.print_exc("Could not get viewport information")

        screen_w, screen_h = self.screen_size()
        print("viewport is", sx, sy, sw, sh)
        if self.use_stretching():
            target_w = screen_w
            target_h = screen_h
        else:
            scale = min(screen_w / sw, screen_h / sh)
            target_w = sw * scale
            target_h = sh * scale
        scale_x = target_w / sw
        scale_y = target_h / sh
        print("scale", scale_x, scale_y)
        orig_scale_x = min(screen_w / ATARI_WIDTH, screen_h / ATARI_HEIGHT)
        orig_scale_y = orig_scale_x
        print("org_scale = ", orig_scale_x, orig_scale_y)
        offset_x = -(sx + (sw / 2.0) - ATARI_WIDTH / 2.0) / \
                    (ATARI_WIDTH / 2.0)
        offset_y = -(sy + (sh / 2.0) - ATARI_HEIGHT / 2.0) / \
                    (ATARI_HEIGHT / 2.0)

        hz = scale_x / orig_scale_x
        vz = scale_y / orig_scale_y
        # hz = int(round(scale_x / orig_scale_x * 100000000))
        # vz = int(round(scale_y / orig_scale_y * 100000000))
        print("horizontal zoom:", hz / 100000000.0)
        print("vertical zoom:", vz / 100000000.0)
        # ho = int(round(offset_x * 100000000))
        # vo = int(round(offset_y * 100000000))
        # print("horizontal offset:", ho / 100000000.0);
        # print("vertical offset:", vo / 100000000.0);
        self.set_env("FILTER_VERT_OFFSET", str(offset_y))
        self.set_env("FILTER_HORIZ_OFFSET", str(offset_x))
        self.set_env("FILTER_VERT_ZOOM", str(vz))
        self.set_env("FILTER_HORIZ_ZOOM", str(hz))

    def atari_model(self):
        return "ST"

    def prepare_tos(self):
        if self.atari_model() == "ST":
            tos_sha1 = TOS_102_UK
        else:
            raise Exception("unknown Atari model")

        uri = self.fsgs.file.find_by_sha1(tos_sha1)
        stream = self.fsgs.file.open(uri)
        self.bios_temp = self.create_temp_file("tos.img")
        with open(self.bios_temp.path, "wb") as f:
            f.write(stream.read())
        return self.bios_temp.path

    def run(self):
        # return self.start_emulator("fs-hatari/hatari")
        return self.start_emulator_from_plugin_resource("hatari")

    def finish(self):
        pass



from game_center.glui.opengl import gl, fs_emu_texturing, fs_emu_blending
from game_center.glui.text import TextRenderer


DISPLAY_FPS = False
DISPLAY_SYNC = False

PERSPECTIVE_STANDARD = 1
PERSPECTIVE_ORTHO = 2
PERSPECTIVE_HD = 3

TEXT_CACHE_SIZE = 100

text_cache_history = []
text_cache = {}


class Render(object):

    currently_ingame = False

    # Render items
    display_sync = DISPLAY_SYNC
    display_fps = DISPLAY_FPS
    display_last_iteration = 0
    display_width = 0
    display_height = 0
    last_time_str = ""

    # State items
    
    time = 0
    down_navigatable = None
    dialog = None
    dialog_time = 0.0
    
    # ...

    quit = False
    allow_minimize = True

    history = []

    zoom = 1.0
    offset_x = 0.0
    offset_y = 0.0

    frame_number = 0
    idle_from = None
    dirty = True
    non_dirty_state = False
    twice = False
    fade_start = 0.0
    fade_end = 0.0
    fade_splash = True
    fade_from = (0, 0, 0, 0)
    fade_to = (0, 0, 0, 0)
    opacity = 1.0
    perspective = 0

    center_item = None
    center_item_time = 0

    items_brightness = 0.8
    # items_brightness_target = 1.0
    # items_brightness_speed = 1.0
    navigatable = None
    current_menu = None
    current_game = None
    config_menu = None
    top_menu = None

    hide_mouse_time = 0
    mouse_visible = True
    # options
    force_portrait = False
    max_ratio = None
    # max_ratio = 1.0

    delete_texture_list = []
    unused_texture_list = []

    @classmethod
    def create_texture(cls):
        if len(cls.unused_texture_list) == 0:
            cls.unused_texture_list.extend(gl.glGenTextures(50))
        return cls.unused_texture_list.pop()

    @classmethod
    def delete_textures(cls):
        # for t in cls.delete_texture_list:
        #     glDeleteTextures([t])

        gl.glDeleteTextures(cls.delete_texture_list)
        cls.delete_texture_list[:] = []

        # cls.unused_texture_list.extend(cls.delete_texture_list)
        # cls.delete_texture_list[:] = []
        # pass

        if len(cls.delete_texture_list) > 0:
            texture = cls.delete_texture_list.pop()
            gl.glDeleteTextures(texture)

    @classmethod
    def standard_perspective(cls):
        if cls.perspective == PERSPECTIVE_STANDARD:
            return
        # aspect_ratio = cls.display_width / cls.display_height
        cls.gl_height = 2.0
        cls.gl_width = 2.0 * cls.display_aspect
        cls.gl_left = -1.0 * cls.display_aspect
        cls.fov_y = 45 / cls.zoom
        cls.fov_x = cls.fov_y * cls.display_aspect / cls.zoom

        # left = cls.offset_x - 1.0 * cls.display_aspect / cls.zoom
        # right = cls.offset_x + 1.0 * cls.display_aspect / cls.zoom
        # top = cls.offset_y + 1.0 / cls.zoom
        # bottom = cls.offset_y - 1.0 / cls.zoom

        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glLoadIdentity()
        
        # gluPerspective(cls.fov_y, cls.display_aspect, 0.1, 10.1)
        # glTranslatef(-cls.offset_x, -cls.offset_y, 0.0)
        
        gl.gluPerspective(45.0, 16.0 / 9.0, 0.1, 100.0)
        
        cls.projection = gl.glGetFloatv(gl.GL_PROJECTION_MATRIX)
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glLoadIdentity()
        # glScissor(0, display_yoffset, cls.display_width, cls.display_height)
        # glEnable(GL_SCISSOR_TEST)
        cls.perspective = PERSPECTIVE_STANDARD

    @classmethod
    def hd_perspective(cls):
        if cls.perspective == PERSPECTIVE_HD:
            return
        cls.display_aspect = 16 / 9.0
        cls.ortho_pscalex = 1920 / cls.display_width
        # cls.ortho_pscalex = 1920 / cls.display_width * \
        #         (cls.display_width / cls.display_height)
        cls.ortho_pscaley = 1080 / cls.display_height
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glLoadIdentity()
        left = 0.0
        right = 1920.0
        top = 1080.0
        bottom = 0.0
        gl.glOrtho(left, right, bottom, top, -1.0, 1.0)
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glLoadIdentity()
        cls.gl_left = left
        cls.gl_right = right
        cls.gl_height = 1080.0
        cls.gl_width = 1920.0
        cls.perspective = PERSPECTIVE_HD
        return cls.ortho_pscalex, cls.ortho_pscaley

    @classmethod
    def ortho_perspective(cls):
        if cls.perspective == PERSPECTIVE_ORTHO:
            return
        cls.display_aspect = 16 / 9.0
        cls.ortho_pscalex = (2.0 / cls.display_width *
                             (cls.display_width / cls.display_height))
        cls.ortho_pscaley = 2.0 / cls.display_height
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glLoadIdentity()
        left = cls.offset_x - 1.0 * cls.display_aspect / cls.zoom
        right = cls.offset_x + 1.0 * cls.display_aspect / cls.zoom
        top = cls.offset_y + 1.0 / cls.zoom
        bottom = cls.offset_y - 1.0 / cls.zoom
        gl.glOrtho(left, right, bottom, top, -1.0, 1.0)
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glLoadIdentity()
        cls.gl_left = -1.0 * cls.display_aspect
        cls.gl_right = 1.0 * cls.display_aspect
        cls.gl_height = 2.0  # * cls.display_aspect
        cls.gl_width = 2.0 * cls.display_aspect
        cls.perspective = PERSPECTIVE_ORTHO
        return cls.ortho_pscalex, cls.ortho_pscaley

    @classmethod
    def text(cls, text, font, x, y, w=0, h=0, color=(1.0, 1.0, 1.0, 1.0),
             shadow=False, shadow_color=(0, 0, 0), halign=-1):
        if not text:
            return 0, 0
        # if len(color) == 3:
        #     color = (color[0], color[1], color[2], 1.0
        try:
            alpha = color[3]
        except IndexError:
            alpha = 1.0
        color = (int(round(color[0] * 255)),
                 int(round(color[1] * 255)),
                 int(round(color[2] * 255)))

        cache_key = (text, hash(font), font.size, color, alpha)
        try:
            text_cache_history.remove(cache_key)
        except ValueError:
            texture = None
        else:
            texture, txtsize = text_cache[cache_key]

        fs_emu_blending(True)
        fs_emu_texturing(True)
        gl.glDisable(gl.GL_DEPTH_TEST)

        if texture:
            gl.glBindTexture(gl.GL_TEXTURE_2D, texture)
        else:
            txtdata, txtsize = TextRenderer(font).render_text(text, color)
            texture = Render.create_texture()
            gl.glBindTexture(gl.GL_TEXTURE_2D, texture)
            gl.glTexImage2D(
                gl.GL_TEXTURE_2D, 0, gl.GL_RGBA, txtsize[0], txtsize[1],
                0, gl.GL_BGRA, gl.GL_UNSIGNED_BYTE, txtdata)
            gl.glTexParameteri(
                gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER, gl.GL_LINEAR)
            gl.glTexParameteri(
                gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER, gl.GL_LINEAR)
        tw, th = txtsize[0] * cls.ortho_pscalex, txtsize[1] * cls.ortho_pscaley

        tx = x
        ty = y
        if w > 0:
            if tw > w:
                tw = w
            else:
                if halign == 0:
                    tx += (w - tw) / 2
        if h > 0:
            ty += (h - th) / 2
        # ts = 4 / cls.display_height  # Step

        # glDisable(GL_TEXTURE_RECTANGLE_ARB)
        
        # glTexEnv(GL_TEXTURE_2D, GL_MODULATE)
        # glDisable(GL_TEXTURE_RECTANGLE)
        # fs_emu_blending(True)
        gl.glBegin(gl.GL_QUADS)
        gl.glColor4f(alpha, alpha, alpha, alpha)

        gl.glTexCoord2f(0.0, 0.0)
        gl.glVertex2f(tx, ty + th)
        gl.glTexCoord2f(1.0, 0.0)
        gl.glVertex2f(tx + tw, ty + th)
        gl.glTexCoord2f(1.0, 1.0)
        gl.glVertex2f(tx + tw, ty)
        gl.glTexCoord2f(0.0, 1.0)
        gl.glVertex2f(tx, ty)

        # glRasterPos2f(tx, ty)
        # glDrawPixels(txtsize[0], txtsize[1], GL_RGBA, GL_UNSIGNED_BYTE, txtdata)
        gl.glEnd()
        
        # fs_emu_blending(False)
        gl.glEnable(gl.GL_DEPTH_TEST)

        text_cache_history.append(cache_key)
        text_cache[cache_key] = texture, txtsize
        if len(text_cache) > TEXT_CACHE_SIZE:
            cache_key = text_cache_history.pop(0)
            texture, txtsize = text_cache.pop(cache_key)
            Render.delete_texture_list.append(texture)

        # # FIXME:
        # shadow = False
        #
        # glDisable(GL_DEPTH_TEST)
        # fs_emu_blending(True)
        # #text = current_menu.selected_item.title
        # #if shadow:
        # txtdata, txtsize = TextRenderer(font).render_text(text, shadow_color)
        # tw, th = txtsize[0] * ortho_pscalex, txtsize[1] * ortho_pscaley
        # tx = x
        # ty = y
        # if w > 0:
        #     tx = tx + (w - tw) / 2
        # if h > 0:
        #     ty = ty + (h - th) / 2
        # #tx = 0 - tw / 2
        # #ty = -0.67
        # ts = 4 / State.display_height # Step
        # if shadow:
        #     glPixelTransferf(GL_ALPHA_SCALE, 0.04)
        #     for fx, fy in [(1, 1), (-1, -1), (1, -1), (-1, 1), (1, 0), (-1, 0),
        #             (0, -1), (0, 1)]:
        #         glRasterPos2f(tx - fx * ts, ty - fy * ts)
        #         glDrawPixels(txtsize[0], txtsize[1], GL_RGBA, GL_UNSIGNED_BYTE,
        #                 txtdata)
        #     glPixelTransferf(GL_ALPHA_SCALE, 0.01)
        #     for fx, fy in [(0, 2), (2, 0), (0, -2), (-2, 0),
        #             (1, 2), (2, 1), (-1, 2), (-2, 1), (1, -2),
        #             (2, -1), (-1, -2), (-2, -1)]:
        #         glRasterPos2f(tx - fx * ts, ty - fy * ts)
        #         glDrawPixels(txtsize[0], txtsize[1], GL_RGBA, GL_UNSIGNED_BYTE,
        #                 txtdata)
        # glPixelTransferf(GL_ALPHA_SCALE, alpha)
        # rendered = font.render(text, True, color)
        # txtsize = rendered.get_size()
        # txtdata = pygame.image.tostring(rendered, "RGBA", 1)
        # glRasterPos2f(tx, ty)
        # glDrawPixels(txtsize[0], txtsize[1], GL_RGBA, GL_UNSIGNED_BYTE, txtdata)
        # #glPopAttrib()
        # glPixelTransferf(GL_ALPHA_SCALE, 1.0)
        # fs_emu_blending(False)
        # glEnable(GL_DEPTH_TEST)
        return tw, th

    @classmethod
    def measure_text(cls, text, font):
        tw, th = TextRenderer(font).get_size(text)
        return tw * cls.ortho_pscalex, th * cls.ortho_pscaley


State = Render



from .qt import QMainWindow
from .windowbase import WindowBase


class Window(WindowBase(QMainWindow)):

    def __init__(self, parent=None, title="", border=True):
        super().__init__(parent, title=title, border=border)


import sys
if "--workspace" in sys.argv:
    fs_uae_workspace = __import__("fs_uae_workspace.window")
    if "--windows" not in sys.argv:
        from fs_uae_workspace.window import Window



from game_center.glui.opengl import *


MAX_LINE = 30


def main(main_window):
    from game_center.glui.settings import Settings

    # for arg in sys.argv:
    #     print("arg -->", arg)
    #     Settings.fullscreen_game = True
    #     if arg == "--fullscreen":
    #         Settings.fullscreen_menu = True
    #     if arg == "--window":
    #         if not "--no-decorations" in sys.argv:
    #             Settings.window_decorations = True
    #         Settings.fullscreen_menu = False
    #         Settings.fullscreen_game = False
    #         # Settings.windowed_size = (1024, 576)
    #     elif arg.startswith("--window="):
    #         Settings.fullscreen_menu = False
    #         Settings.fullscreen_game = False
    #         w, h = arg[9:].split("x")
    #         w = int(w)
    #         h = int(h)
    #         Settings.windowed_size = (w, h)
    #     elif arg == "--maximize":
    #         if not "--no-decorations" in sys.argv:
    #             Settings.window_decorations = True
    #         Settings.fullscreen_menu = False
    #         Settings.fullscreen_game = True
    #         Settings.windowed_size = None
    #     elif arg == "--no-decorations":
    #         Settings.window_decorations = False
    #     elif arg == "--decorations":
    #         Settings.window_decorations = True

    try:
        import game_center.glui.window
        game_center.glui.window.main_window = main_window
        game_center.glui.window.show()
    except Exception as e:
        import traceback
        print("\n" + "-" * 79 + "\n" + "EXCEPTION")
        traceback.print_exc()
        show_error(repr(e), traceback.format_exc(e))

    print(" --- game_center.glui.main is done ---")


# import pygame
# from game_center.glui.render import Render
#
#
# def show_error(message, backtrace):
#     width = 16 / 9 * 2
#     height = 2.0
#     message = message
#     backtrace = backtrace
#     splitted = backtrace.split("\n")
#     if not splitted[-1]:
#         splitted = splitted[:-1]
#
#     #background_color = (0.0, 0.0, 0.0, 1.0)
#     liberation_mono_bold = resources.resource_filename(
#         "LiberationMono-Regular.ttf")
#     print(liberation_mono_bold)
#     detail_font = pygame.font.Font(
#         liberation_mono_bold, int(0.021 * Render.display_height))
#     guru_font = pygame.font.Font(
#         liberation_mono_bold, int(0.03 * Render.display_height))
#     start_time = time.time()
#     while True:
#         events = pygame.event.get()
#         stop = False
#         for event in events:
#             if event.type == pygame.QUIT:
#                 stop = True
#             if event.type == pygame.KEYDOWN:
#                 if event.key == pygame.K_RETURN:
#                     stop = True
#                 if event.key == pygame.K_ESCAPE:
#                     stop = True
#         if stop:
#             pygame.display.quit()
#             sys.exit(1)
#
#         glClearColor(0.0, 0.0, 0.0, 0.0)
#         glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
#         Render.ortho_perspective()
#         glPushMatrix()
#         glTranslatef(-width / 2, -height / 2, 0.0)
#         x1 = 0
#         x2 = width
#         y1 = 1.7
#         y2 = 2.0
#         w = 0.03
#
#         glDisable(GL_DEPTH_TEST)
#         fs_emu_texturing(False)
#         fs_emu_blending(True)
#         alert_color = (1.0, 0.0, 0.0)
#         t = int((time.time() - start_time) * 1.6)
#         if t % 2 == 0:
#             glBegin(GL_QUADS)
#             glColor3f(*alert_color)
#             glVertex2f(x1, y1)
#             glVertex2f(x2, y1)
#             glVertex2f(x2, y2)
#             glVertex2f(x1, y2)
#             glColor3f(0.0, 0.0, 0.0)
#             glVertex2f(x1 + w, y1 + w)
#             glVertex2f(x2 - w, y1 + w)
#             glVertex2f(x2 - w, y2 - w)
#             glVertex2f(x1 + w, y2 - w)
#             glEnd()
#         fs_emu_texturing(True)
#         glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA)
#
#         text = "Software Failure.  Press ENTER or ESCAPE key to quit."
#         Render.text(text, guru_font, 0.2, 1.85, color=alert_color)
#         text = splitted[-1]
#         text = "Guru Meditation #{0}".format(text)
#         Render.text(text, guru_font, 0.2, 1.77, color=alert_color)
#
#         x = 0.2
#         y = 0.15
#
#         tw, th = Render.measure_text("M", detail_font)
#         y += th
#         lines = []
#         max_line_size = 129
#         for line in splitted:
#             line = line.rstrip()
#             while len(line) > max_line_size:
#                 lines.append(line[:max_line_size])
#                 line = line[max_line_size:]
#             lines.append(line)
#
#         for i, line in enumerate(reversed(lines)):
#             if i == MAX_LINE:
#                 break
#             s = (MAX_LINE - i) / MAX_LINE
#             tw, th = Render.text(
#                 line, detail_font, x, y, color=(s, s, s, 1.0))
#             y += th
#         glPopMatrix()
#         pygame.display.flip()



from fs_uae_launcher.Options import Option
from fs_uae_launcher.ui.config.configpanel import ConfigPanel
from fs_uae_launcher.ui.Skin import Skin
# import fsui as fsui
from .KickstartGroup import KickstartGroup
from .MemoryGroup import MemoryGroup
# from .config.ExpansionsGroup import ExpansionsGroup


class RomRamPanel(ConfigPanel):

    def __init__(self, parent):
        ConfigPanel.__init__(self, parent)

        self.kickstart_group = KickstartGroup(self)
        self.memory_group = MemoryGroup(self)
        # self.expansions_group = ExpansionsGroup(panel)

        self.layout.add(self.kickstart_group, fill=True)
        self.layout.add_spacer(10)
        self.layout.add_spacer(Skin.EXTRA_GROUP_MARGIN)
        self.layout.add(self.memory_group, fill=True)

        self.layout.add_spacer(-10)

        # self.add_option(Option.ACCELERATOR_MEMORY)

        # panel.layout.add_spacer(Skin.EXTRA_GROUP_MARGIN)
        # panel.layout.add(self.expansions_group, fill=True)



import fsui as fsui
from .Skin import Skin


class InputPanel(fsui.Panel):

    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        Skin.set_background_color(self)

        self.layout = fsui.VerticalLayout()

        from .config.InputGroup import InputGroup
        self.input_group = InputGroup(self, with_more_options=True)
        self.layout.add(self.input_group, fill=True)
        self.layout.add_spacer(Skin.EXTRA_GROUP_MARGIN)

        self.input_group = InputGroup(self, with_more_options=False,
                                      parallel_ports=True)
        self.layout.add(self.input_group, fill=True)

        self.input_group = InputGroup(self, with_more_options=False,
                                      custom_ports=True)
        self.layout.add(self.input_group, fill=True)



from game_center.glui.opengl import glDeleteLists, glCallList, glNewList, glGenLists
from game_center.glui.opengl import GL_COMPILE_AND_EXECUTE


class DisplayLists(object):

    display_lists = {}

    @classmethod
    def clear(cls):
        for listid in cls.display_lists.values():
            glDeleteLists(listid, 1)
        cls.display_lists.clear()

    @classmethod
    def call_or_create(cls, *params):
        try:
            glCallList(cls.display_lists[params])
            return False
        except KeyError:
            display_list = glGenLists(1)
            print("creating display list for params", params)
            glNewList(display_list, GL_COMPILE_AND_EXECUTE)
            cls.display_lists[params] = display_list
            return True



'''OpenGL extension INGR.blend_func_separate

This module customises the behaviour of the 
OpenGL.raw.GL.INGR.blend_func_separate to provide a more 
Python-friendly API

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/INGR/blend_func_separate.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
from OpenGL.GL import glget
import ctypes
from OpenGL.raw.GL import _types
from OpenGL.raw.GL.INGR.blend_func_separate import *
from OpenGL.raw.GL.INGR.blend_func_separate import _EXTENSION_NAME

def glInitBlendFuncSeparateINGR():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )

### END AUTOGENERATED SECTION


from urllib.request import urlopen
import traceback
import threading
from .Signal import Signal
from .Settings import Settings
from fsbc.Application import Application
from fsbc.system import windows, macosx, linux
from fsbc.util import compare_versions, unused
from fstd.desktop import open_url_in_browser


class UpdateManager:

    @classmethod
    def run_update_check(cls):
        threading.Thread(target=cls.update_thread_function).start()

    @classmethod
    def update_thread_function(cls):
        try:
            cls._update_thread_function()
        except Exception:
            traceback.print_exc()

    @classmethod
    def series(cls):
        if "dev" in Application.instance().version:
            return "devel"
        elif "beta" in Application.instance().version:
            return "beta"
        return "stable"

    @classmethod
    def _update_thread_function(cls):
        if windows:
            platform = "windows"
        elif macosx:
            platform = "macosx"
        elif linux:
            platform = "linux"
        else:
            platform = "other"
        url = "http://fs-uae.net/{0}/latest-{1}".format(cls.series(), platform)
        f = urlopen(url)
        version_str = f.read().strip().decode("UTF-8")
        print("latest version available: ", version_str)
        print("current version: ", Application.instance().version)
        result = compare_versions(version_str, Application.instance().version)
        print("update check result: ", result)
        if result > 0 and version_str != "9.9.9":
            web_url = "http://fs-uae.net/{0}/download/".format(cls.series())
            Signal.broadcast("update_available", version_str, web_url)
            Settings.set("__update_available", version_str)

    @classmethod
    def start_update(cls, version_str):
        unused(version_str)
        web_url = "http://fs-uae.net/{0}/download/".format(cls.series())
        open_url_in_browser(web_url)



import os
from fsgs.context import fsgs
from fsgs.ChecksumTool import ChecksumTool
import fsui as fsui
from fsgs.amiga.Amiga import Amiga
from ..CDManager import CDManager
from ..Config import Config
from ..FloppyManager import FloppyManager
from ..I18N import gettext
from fsbc.Paths import Paths
from fsgs.FSGSDirectories import FSGSDirectories
from .IconButton import IconButton
from .LauncherFilePicker import LauncherFilePicker


class MediaListGroup(fsui.Group):

    def __init__(self, parent, cd_mode):
        fsui.Group.__init__(self, parent)
        self.layout = fsui.VerticalLayout()

        self.cd_mode = cd_mode
        if self.cd_mode:
            self.file_key_prefix = "cdrom_image_"
            self.file_key = "cdrom_image_{0}"
            self.sha1_key = "x_cdrom_image_{0}_sha1"
        else:
            self.file_key_prefix = "floppy_image_"
            self.file_key = "floppy_image_{0}"
            self.sha1_key = "x_floppy_image_{0}_sha1"

        self.heading_label = fsui.HeadingLabel(
            self, gettext("Media Swap List"))
        self.layout.add(self.heading_label, margin=10)
        self.layout.add_spacer(0)

        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, expand=True, fill=True)

        self.list_view = fsui.ListView(self)
        self.list_view.on_activate_item = self.on_activate_item
        if self.cd_mode:
            # self.list_view.set_default_icon(
            #     fsui.Image("fs_uae_launcher:res/cdrom_16.png"))
            self.default_icon = fsui.Image("fs_uae_launcher:res/cdrom_16.png")
        else:
            # self.list_view.set_default_icon(
            #     fsui.Image("fs_uae_launcher:res/floppy_16.png"))
            self.default_icon = fsui.Image("fs_uae_launcher:res/floppy_16.png")
        hori_layout.add(self.list_view, expand=True, fill=True, margin=10,
                        margin_right=0)

        vert_layout = fsui.VerticalLayout()
        hori_layout.add(vert_layout, fill=True)

        add_button = IconButton(self, "add_button.png")
        add_button.set_tooltip(gettext("Add Files to List"))
        add_button.activated.connect(self.on_add_button)
        vert_layout.add(add_button, margin=10)

        remove_button = IconButton(self, "remove_button.png")
        remove_button.set_tooltip(gettext("Remove Selected Files"))
        remove_button.activated.connect(self.on_remove_button)
        vert_layout.add(remove_button, margin=10)

        clear_button = IconButton(self, "clear_button.png")
        clear_button.set_tooltip(gettext("Clear List"))
        clear_button.activated.connect(self.on_clear_list)
        vert_layout.add(clear_button, margin=10)

        self.update_list()
        Config.add_listener(self)

    def on_destroy(self):
        Config.remove_listener(self)

    def on_config(self, key, _):
        if key.startswith(self.file_key_prefix):
            self.update_list()

    def on_activate_item(self, item):
        path = Config.get(self.file_key.format(item))
        sha1 = Config.get(self.sha1_key.format(item))
        if self.cd_mode:
            pass
        else:
            fsgs.amiga.insert_floppy_in_free_drive(path, sha1=sha1)

    def create_list(self):
        items = []
        if self.cd_mode:
            max_items = Amiga.MAX_CDROM_IMAGES
        else:
            max_items = Amiga.MAX_FLOPPY_IMAGES
        for i in range(max_items):
            path = Config.get(self.file_key.format(i))
            sha1 = Config.get(self.sha1_key.format(i))
            if not path:
                continue
            items.append((path, sha1))
        return items

    def update_list(self):
        # items = []
        self.list_view.clear()
        for path, sha1 in self.create_list():
            dir, name = os.path.split(path)
            if dir:
                label = "{0} ({1})".format(name, dir)
            else:
                label = path
            self.list_view.add_item(label, self.default_icon)
        # self.list_view.set_items(items)

    def on_clear_list(self):
        if self.cd_mode:
            CDManager.clear_cdrom_list()
        else:
            FloppyManager.clear_floppy_list()

    def on_remove_button(self):
        index = self.list_view.get_index()
        existing_items = self.create_list()
        if 0 <= index < len(existing_items):
            del existing_items[index]
            self.set_new_config(existing_items)

    def on_add_button(self):
        existing_items = self.create_list()

        default_dir = FSGSDirectories.get_floppies_dir()
        if self.cd_mode:
            dialog = LauncherFilePicker(
                self.get_window(), gettext("Select Multiple CD-ROMs"),
                "cd", multiple=True)
        else:
            dialog = LauncherFilePicker(
                self.get_window(), gettext("Select Multiple Floppies"),
                "floppy", multiple=True)
        if not dialog.show_modal():
            print("dialog.show returned false")
            return
        print("dialog.show returned true")
        paths = dialog.get_paths()
        paths.sort()
        print(paths)

        checksum_tool = ChecksumTool(self.get_window())
        for i, path in enumerate(paths):
            sha1 = checksum_tool.checksum(path)
            path = Paths.contract_path(path, default_dir)

            dir, file = os.path.split(path)
            if os.path.normcase(os.path.normpath(dir)) == \
                    os.path.normcase(os.path.normpath(default_dir)):
                path = file

            existing_items.append((path, sha1))
        self.set_new_config(existing_items)

    def set_new_config(self, items):
        if self.cd_mode:
            max_items = Amiga.MAX_CDROM_IMAGES
        else:
            max_items = Amiga.MAX_FLOPPY_IMAGES
        set_list = []
        for i in range(max(max_items, len(items))):
            if i >= max_items:
                break
            elif i >= len(items):
                path, sha1 = "", ""
            else:
                path, sha1 = items[i]
            set_list.append((self.file_key.format(i), path))
            set_list.append((self.sha1_key.format(i), sha1))
        Config.set_multiple(set_list)



import sys
import threading

from fsbc.Application import call_after
from ..Settings import Settings
from oyoyo.client import IRCClient
from oyoyo import helpers
from .CommandHandler import CommandHandler
from .IRCBroadcaster import IRCBroadcaster
from .IRCColor import IRCColor


# noinspection PyUnusedLocal
class IRC:

    running = False
    stopping = False
    nick_number = 0
    channels = {}
    my_nick = ""
    client = None

    # FIXME: should not be hardcoded here
    default_channel = "#lobby"

    active_channel = default_channel

    @classmethod
    def message(cls, message, color=None):
        cls.channel(cls.active_channel).message(message, color)

    @classmethod
    def warning(cls, message):
        cls.message(message, IRCColor.WARNING)

    @classmethod
    def info(cls, message):
        cls.message(message, IRCColor.INFO)

    @classmethod
    def is_channel(cls, name):
        return len(name) > 0 and name[0] in "&#+!"

    @classmethod
    def channel(cls, name):
        # FIXME: name should be checked case independently
        if not name:
            name = cls.default_channel

        from .Channel import Channel
        try:
            return cls.channels[name]
        except KeyError:
            print("new channel", repr(name))
            cls.channels[name] = Channel(name)
            print("channels are now", repr(cls.channels))
            # cls.set_active_channel(name)
            IRCBroadcaster.broadcast("channel_list", {"added": name})
            return cls.channels[name]

    @classmethod
    def set_active_channel(cls, name):
        cls.active_channel = name
        IRCBroadcaster.broadcast("active_channel", {"channel": name})

    @classmethod
    def connect(cls):
        cls.start()

    @classmethod
    def start(cls):
        cls.stopping = False
        if cls.running:
            print("IRC.start - already running")
            return
        threading.Thread(target=cls.irc_thread,
                         name="IRCThread").start()
        cls.running = True

    @classmethod
    def stop(cls):
        if not cls.running:
            return
        cls.stopping = True
        helpers.quit(cls.client, "Exited")
        # cls.client.quit()

    @classmethod
    def irc_thread(cls):
        try:
            cls.irc_main()
        except Exception as e:
            def func():
                cls.warning(repr(e))
            call_after(func)
            import traceback
            traceback.print_exc()
        cls.running = False

    @classmethod
    def nick(cls, spec):
        nick = spec.split("!")[0]
        nick = nick.strip("@+")
        return nick

    @classmethod
    def me(cls, spec):
        nick = spec.split("!")[0]
        return nick == cls.my_nick

    @classmethod
    def get_irc_server_host(cls):
        return Settings.get_irc_server()

    @classmethod
    def irc_main(cls):
        def func():
            cls.message("connecting to {0}...".format(
                cls.get_irc_server_host()))
        call_after(func)

        cls.client = IRCClient(
            CommandHandler,
            host=cls.get_irc_server_host(),
            port=6667,
            nick=cls.generate_nick(True),
            blocking=True,
            connect_cb=cls.connect_callback)
        cls.client.handler = cls
        cls.connection = cls.client.connect()

        while not cls.stopping:
            next(cls.connection)
        print("irc_main done")
        cls.running = False

    @classmethod
    def connect_callback(cls, client):
        def func():
            cls.message("connected to {0}".format(cls.get_irc_server_host()))
        call_after(func)

    @classmethod
    def post_message(cls, command, args):
        # command = command.decode("UTF-8", errors="replace")
        if cls.stopping:
            print("ignoring message", command, "because we're stopping")
            return
        # print("post_message", command, args)
        # cls.on_message(command, args)
        # print(command, args)

        args = list(args)
        # for i, arg in enumerate(args):
        #     if arg is not None:
        #         args[i] = arg.decode("UTF-8", errors="replace")

        def func():
            # if IRC.broadcast(command, args):
            #     return
            if cls.irc_server_message(command, args):
                return
            name = "irc_" + command
            try:
                method = getattr(cls, name)
            except AttributeError:
                cls.info(" ".join([command] + args))
                pass
            else:
                method(*args)

        call_after(func)

    @classmethod
    def privmsg(cls, destination, message):
        helpers.msg(cls.client, destination, message)

    @classmethod
    def notice(cls, destination, message):
        cls.client.send("notice {0} :{1}".format(destination, message))

    @classmethod
    def generate_nick(cls, reset=False):
        if reset:
            cls.nick_number = 0
        cls.nick_number += 1
        if cls.nick_number == 1:
            nick = Settings.get_irc_nick()
        else:
            nick = Settings.get_irc_nick() + str(cls.nick_number)
        cls.my_nick = nick
        return nick

    @classmethod
    def append_text(cls, message):
        cls.message(message)

    @classmethod
    def handle_command(cls, message):
        if message.startswith("/"):
            return cls.handle_command_string(message)
        else:
            if not cls.active_channel:
                cls.warning("no active channel")
            else:
                cls.channel(cls.active_channel).privmsg(message)

    @classmethod
    def handle_command_string(cls, message):
        message = message[1:]
        parts = message.split(" ")
        command = parts[0].lower()
        args = parts[1:]
        name = "command_" + command
        try:
            method = getattr(cls, name)
        except AttributeError:
            cls.warning(command + ": unknown command")
            return False
        else:
            method(args)
            return True

    @classmethod
    def command_raw(cls, args):
        if len(args) >= 1:
            cls.client.send(" ".join(args))
        else:
            cls.warning("usage: /raw <raw irc message>")

    @classmethod
    def command_whois(cls, args):
        if len(args) >= 1:
            cls.client.send("whois {0}".format(" ".join(args)))
        else:
            cls.warning("usage: /whois <nick>")

    @classmethod
    def command_away(cls, args):
        if len(args) == 0:
            cls.client.send("away")
        else:
            cls.client.send("away {0}".format(" ".join(args)))

    @classmethod
    def command_back(cls, args):
        if len(args) == 0:
            cls.client.send("away")
        else:
            cls.warning("usage: /back")

    @classmethod
    def command_msg(cls, args):
        if len(args) >= 2:
            channel = args[0]
            message = " ".join(args[1:])
            # cls.channel(channel).privmsg(message)
            # cls.channel(channel).message("<{0}> {1}".format(cls.my_nick,
            #        message), IRCColor.MY_MESSAGE)
            cls.client.send("privmsg {0} :{1}".format(channel, message))
        else:
            cls.warning("usage: /msg <nick|channel> <message>")

    @classmethod
    def command_notice(cls, args):
        if len(args) >= 2:
            channel = args[0]
            message = " ".join(args[1:])
            # cls.channel(channel).notice(message)
            cls.client.send("notice {0} :{1}".format(channel, message))
        else:
            cls.warning("usage: /notice <nick|channel> <message>")

    @classmethod
    def command_oper(cls, args):
        if len(args) == 2:
            cls.client.send("oper {0} {1}".format(args[0], args[1]))
        else:
            cls.warning("usage: /oper <user> <password>")

    @classmethod
    def command_slap(cls, args):
        if len(args) == 1:
            message = "slaps {0} around a bit with a large trout".format(
                args[0])
            cls.channel(cls.active_channel).action(message)
        else:
            cls.warning("usage: /slap <nick>")

    @classmethod
    def command_me(cls, args):
        if len(args) > 0:
            message = " ".join(args)
            cls.channel(cls.active_channel).action(message)
        else:
            cls.warning("usage: /me <message>")

    @classmethod
    def command_mode(cls, args):
        if len(args) >= 2:
            cls.client.send("mode {0}".format(" ".join(args)))
        else:
            cls.warning("usage: /mode <channel|nick> <parameters...>")

    @classmethod
    def command_kick(cls, args):
        if not cls.is_channel(cls.active_channel):
            cls.warning("cannot kick - not in a channel")
            return
        if len(args) >= 1:
            cls.kick(cls.active_channel, args[0], " ".join(args[1:]))
        else:
            cls.warning("usage: /kick <nick> [<message>]")

    @classmethod
    def kick(cls, channel, nick, message="kicked"):
        cls.client.send("KICK", channel, nick, ":{0}".format(message))

    @classmethod
    def command_join(cls, args):
        if len(args) == 1:
            cls.join(args[0])
        else:
            cls.warning("join: invalid arguments")

    @classmethod
    def join(cls, channel):
        helpers.join(cls.client, channel)

    @classmethod
    def command_leave(cls, args):
        return cls.command_part(args)

    @classmethod
    def command_part(cls, args):
        if len(args) == 1:
            cls.part(args[0])
        else:
            cls.warning("part: invalid arguments")

    @classmethod
    def part(cls, channel):
        helpers.part(cls.client, channel)

    @classmethod
    def irc_server_message(cls, command, args):
        if command in [
                "yourhost", "created", "luserclient", "luserchannels",
                "luserme", "n_local", "n_global", "luserconns", "motdstart",
                "motd", "nomotd", "endofmotd"]:
            cls.message(args[2])
            return 1
        return 0

    @classmethod
    def irc_featurelist(cls, server, nick, *features):
        cls.message(" ".join(features))

    @classmethod
    def irc_welcome(cls, server, nick, message):
        cls.message(message)

        password = Settings.get_irc_nickserv_pass()
        if password:
            cls.privmsg("nickserv", "identify {0}".format(password))

        # cls.join("#support")
        cls.join("#lobby")

        # convenience for development...
        for arg in sys.argv:
            if arg.startswith("--netplay-game="):
                cls.join("&" + arg[15:])
                break

    @classmethod
    def filter_nick(cls, full):
        nick = full.split("!")[0]
        return nick

    @classmethod
    def irc_privmsg(cls, sender, destination, message):
        nick = cls.filter_nick(sender)
        if cls.is_channel(destination):
            channel = destination
        else:
            channel = IRC.filter_nick(sender)
        cls.channel(channel).on_privmsg(nick, message)
        IRCBroadcaster.broadcast("privmsg", {
            "channel": channel, "message": message, "nick": nick})

    @classmethod
    def irc_notice(cls, sender, destination, message):
        if sender:
            nick = cls.filter_nick(sender)
        else:
            nick = "(server)"
        if cls.is_channel(destination):
            channel = destination
        else:
            channel = cls.active_channel
        cls.channel(channel).on_notice(nick, message)
        IRCBroadcaster.broadcast("notice", {
            "channel": channel, "message": message, "nick": nick})

    @classmethod
    def irc_kick(cls, kicker, channel, who, why):
        cls.channel(channel).on_kick(IRC.filter_nick(kicker), who, why)

    @classmethod
    def irc_part(cls, who, channel, *_):
        cls.channel(channel).on_part(IRC.filter_nick(who))

    @classmethod
    def irc_join(cls, who, channel):
        cls.channel(channel).on_join(IRC.filter_nick(who))

    @classmethod
    def irc_endofnames(cls, sender, recipient, channel, message):
        pass

    @classmethod
    def irc_namreply(cls, server, who, equals, channel, nicks):
        cls.channel(channel).on_namreply(nicks.split(" "))

    @classmethod
    def irc_currenttopic(cls, sender, recipient, channel, topic):
        cls.channel(channel).on_currenttopic(topic)

    @classmethod
    def irc_topicinfo(cls, sender, recipient, channel, who, when):
        pass

    @classmethod
    def irc_cannotsendtochan(cls, server, who, channel, message):
        cls.channel(channel).warning("cannot send to channel: " + message)

    @classmethod
    def irc_topic(cls, who, channel, topic):
        cls.channel(channel).on_topic(who, topic)

    @classmethod
    def irc_mode(cls, who, destination, *args):
        if cls.is_channel(destination):
            cls.channel(destination).on_mode(IRC.filter_nick(who), args)
        else:
            cls.message("{0} set mode {1} {2}".format(
                who, destination, " ".join(args)))

    @classmethod
    def irc_quit(cls, who, reason):
        for name in cls.channels:
            cls.channel(name).on_quit(IRC.filter_nick(who), reason)



from fsbc.Application import app
from fsbc.user import get_user_name
from .Signal import Signal


class Settings(object):
    irc_server = "irc.fengestad.no"

    base_dir = ""

    # The main use of default_settings is actually to hide the corresponding
    # options from advanced settings (because there is UI for them, or
    # because the settings are internal)
    default_settings = {
        "__netplay_ready": "",
        "automatic_input_grab": "",
        "audio_frequency": "",
        "audio_buffer_target_size": "",
        "builtin_configs": "",
        "config_base": "",
        "config_changed": "0",
        "config_name": "Unnamed Configuration",
        "config_feature": "",
        "config_path": "",
        "config_refresh": "",
        "config_search": "",
        "config_xml_path": "",
        "configurations_dir_mtime": "",
        "database_auth": "",
        "database_email": "",
        "database_feature": "",
        "database_password": "",
        "database_server": "",
        "database_show_adult": "",
        "database_show_games": "",
        "database_username": "",
        "device_id": "",
        "fsaa": "",
        "floppy_drive_volume": "",
        "floppy_drive_volume_empty": "",
        "fullscreen": "",
        "game_uuid": "",
        "game_list_uuid": "",
        "initial_input_grab": "",
        "irc_nick": "",
        "irc_server": "",
        "keep_aspect": "",
        "kickstarts_dir_mtime": "",
        "kickstart_setup": "",
        "language": "",
        "last_cd_dir": "",
        "last_floppy_dir": "",
        "last_hd_dir": "",
        "last_rom_dir": "",
        "last_scan": "",
        "last_settings_page": "",
        "launcher_theme": "",
        "low_latency_vsync": "",
        "maximized": "0",
        "middle_click_ungrab": "",
        "monitor": "",
        "mouse_speed": "",
        "netplay_feature": "",
        "netplay_tag": "",
        "parent_uuid": "",
        "primary_joystick": "",
        # "rtg_scanlines": "",
        "scan_configs": "1",
        "scan_files": "1",
        "scan_roms": "1",
        # "scanlines": "",
        "search_path": "",
        "secondary_joystick": "",
        "stereo_separation": "",
        "swap_ctrl_keys": "",
        "texture_filter": "",
        "texture_format": "",
        "video_format": "",
        "video_sync": "",
        "video_sync_method": "",
        "zoom": "",
        # "window_width": "",
        # "window_height": "",
    }

    settings = default_settings.copy()
    # settings_listeners = []

    initialize_from_config = set([
        "fullscreen",
    ])

    @classmethod
    def get(cls, key):
        return app.settings[key]

    @classmethod
    def set(cls, key, value):
        app.settings[key] = value

    @classmethod
    def add_listener(cls, listener):
        # cls.settings_listeners.append(listener)
        Signal.add_listener("setting", listener)

    @classmethod
    def remove_listener(cls, listener):
        # cls.settings_listeners.remove(listener)
        Signal.remove_listener("setting", listener)

    @classmethod
    def get_irc_nick(cls):
        value = cls.get("irc_nick").strip()
        if not value:
            value = get_user_name()
        # these are probably valid too: \`
        valid_chars = ("ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                       "abcdefghijklmnopqrstuvwxyz"
                       "_[]{}|^")
        extra_valid_chars = "0123456789-"
        nick = ""
        for c in value:
            if c in valid_chars:
                nick = nick + c
                if extra_valid_chars:
                    valid_chars += extra_valid_chars
                    extra_valid_chars = ""
        if not nick:
            nick = "User"
        return nick

    @classmethod
    def get_irc_nickserv_pass(cls):
        value = cls.get("irc_nickserv_pass").strip()
        if value:
            return value
        return ""

    @classmethod
    def get_irc_server(cls):
        value = cls.get("irc_server").strip()
        if value:
            return value
        return "irc.fengestad.no"



from .mess import MESSRunner


class Atari7800Runner(MESSRunner):

    CONTROLLER = {
        "type": "controller",
        "description": "Controller",
        "mapping_name": "atari7800",
    }

    PORTS = [
        {
            "description": "1st Controller",
            "types": [CONTROLLER]
        }, {
            "description": "2nd Controller",
            "types": [CONTROLLER]
        },
    ]

    def mess_configure(self):
        self.mess_configure_cartridge()

    def mess_input_mapping(self, port):
        return {
            "START": "P#_START",
            "UP": "P#_JOYSTICK_UP",
            "DOWN": "P#_JOYSTICK_DOWN",
            "LEFT": "P#_JOYSTICK_LEFT",
            "RIGHT": "P#_JOYSTICK_RIGHT",
            "1": "P#_BUTTON1",
            "2": "P#_BUTTON2",
        }

    def mess_romset(self):
        if self.is_pal():
            return "a7800p", A7800P_ROMS
        else:
            return "a7800", A7800_ROMS


A7800_ROMS = {
    "d9d134bb6b36907c615a594cc7688f7bfcef5b43": "7800.u7",
    "14584b1eafe9721804782d4b1ac3a4a7313e455f": "c300558-001a.u7",
}


A7800P_ROMS = {
    "5a140136a16d1d83e4ff32a19409ca376a8df874": "7800pal.rom",
}



import fsui as fsui
from fsui.extra.iconheader import NewIconHeader
from ...I18N import gettext
from .CustomSettingsPage import CustomSettingsPage


class AdvancedSettingsPage(fsui.Panel):

    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        self.layout = fsui.VerticalLayout()
        # self.layout.set_padding(20, 20, 20, 20)

        self.icon_header = NewIconHeader(
            self, fsui.Icon("settings", "pkg:fs_uae_workspace"),
            gettext("Advanced Settings"),
            gettext("Specify global options and settings which does "
                    "not have UI controls"))
        self.layout.add(self.icon_header, fill=True, margin_bottom=20)
        self.layout.add(CustomSettingsPage(self), fill=True, expand=True)



import os
import traceback
from fsgs.ChecksumTool import ChecksumTool
import fsui as fsui
from fsgs.Archive import Archive
from ...Config import Config
from fsgs.FSGSDirectories import FSGSDirectories
from ...I18N import gettext
from ..IconButton import IconButton
from ..LauncherFilePicker import LauncherFilePicker


class HardDriveGroup(fsui.Group):

    def __init__(self, parent, index):
        fsui.Group.__init__(self, parent)
        self.layout = fsui.VerticalLayout()

        self.index = index
        self.config_key = "hard_drive_{0}".format(index)
        self.config_key_sha1 = "x_hard_drive_{0}_sha1".format(index)

        if index == 0:
            # heading_label = fsui.HeadingLabel(self,
            #         _("Hard Drive {0}").format(index + 1))
            heading_label = fsui.HeadingLabel(self, gettext("Hard Drives"))
            self.layout.add(heading_label, margin_bottom=20)
            self.layout.add_spacer(0)

        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)

        self.eject_button = IconButton(self, "eject_button.png")
        self.eject_button.set_tooltip(gettext("Eject"))
        self.eject_button.activated.connect(self.on_eject_button)
        hori_layout.add(self.eject_button)

        self.text_field = fsui.TextField(self, "", read_only=True)
        hori_layout.add(self.text_field, expand=True, margin_left=10)

        self.browse_button = IconButton(self, "browse_folder_16.png")
        self.browse_button.set_tooltip(gettext("Browse for Folder"))
        self.browse_button.activated.connect(self.on_browse_folder_button)
        hori_layout.add(self.browse_button, margin_left=10)

        self.browse_button = IconButton(self, "browse_file_16.png")
        self.browse_button.set_tooltip(gettext("Browse for File"))
        self.browse_button.activated.connect(self.on_browse_file_button)
        hori_layout.add(self.browse_button, margin_left=10)

        self.initialize_from_config()
        self.set_config_handlers()

    def initialize_from_config(self):
        self.on_config(self.config_key, Config.get(self.config_key))

    def set_config_handlers(self):
        Config.add_listener(self)

    def on_destroy(self):
        Config.remove_listener(self)

    def on_config(self, key, value):
        if key != self.config_key:
            return
        dir, name = os.path.split(value)
        if dir:
            path = "{0} ({1})".format(name, dir)
        else:
            path = name
        self.text_field.set_text(path)

    def on_eject_button(self):
        Config.set_multiple([(self.config_key, ""),
                             (self.config_key_sha1, "")])

    def on_browse_folder_button(self):
        self.browse(dir_mode=True)

    def on_browse_file_button(self):
        self.browse(dir_mode=False)

    def browse(self, dir_mode):
        default_dir = FSGSDirectories.get_hard_drives_dir()
        dialog = LauncherFilePicker(
            self.get_window(), gettext("Choose Hard Drive"), "hd",
            Config.get(self.config_key), dir_mode=dir_mode)
        if not dialog.show_modal():
            dialog.destroy()
            return
        path = dialog.get_path()
        dialog.destroy()

        checksum_tool = ChecksumTool(self.get_window())
        sha1 = ""
        if dir_mode:
            print("not calculating HD checksums for directories")
        else:
            size = os.path.getsize(path)
            if size < 64 * 1024 * 1024:
                sha1 = checksum_tool.checksum(path)
            else:
                print("not calculating HD checksums HD files > 64MB")
        full_path = path

        # FIXME: use contract function
        dir, file = os.path.split(path)
        self.text_field.set_text(file)
        if os.path.normcase(os.path.normpath(dir)) == \
                os.path.normcase(os.path.normpath(default_dir)):
            path = file

        self.text_field.set_text(path)
        values = [(self.config_key, path),
                  (self.config_key_sha1, sha1)]
        if self.index == 0:
            # whdload_args = ""
            # dummy, ext = os.path.splitext(path)
            # if not dir_mode and ext.lower() in Archive.extensions:
            #     try:
            #         whdload_args = self.calculate_whdload_args(full_path)
            #     except Exception:
            #         traceback.print_exc()
            # values.append(("x_whdload_args", whdload_args))
            values.extend(self.generate_config_for_archive(
                full_path, model_config=False).items())
        Config.set_multiple(values)

    @classmethod
    def generate_config_for_archive(cls, path, model_config=True):
        values = {}
        whdload_args = ""
        dummy, ext = os.path.splitext(path)
        if ext.lower() in Archive.extensions:
            try:
                whdload_args = cls.calculate_whdload_args(path)
            except Exception:
                traceback.print_exc()
        values["x_whdload_args"] = whdload_args
        if whdload_args and model_config:
            values["amiga_model"] = "A1200"
            values["fast_memory"] = "8192"
        return values

    @classmethod
    def calculate_whdload_args(cls, archive_path):
        archive = Archive(archive_path)
        slave = ""
        for path in archive.list_files():
            name = os.path.basename(path)
            name_lower = name.lower()
            if name_lower.endswith(".slave"):
                if slave:
                    print("already found one slave, don't know which "
                          "one to choose")
                    return ""
                slave = name
            elif name_lower == "startup-sequence":
                print("found startup-sequence, assuming non-whdload "
                      "archive")
        return slave



from fsbc.util import unused
import fsui as fsui
from ...Settings import Settings
from ...GamePaths import GamePaths
from ..Constants import Constants
from ..ImageLoader import ImageLoader
from ..Skin import Skin
from .BottomPanel import BottomPanel

BORDER = 20


class ScreenshotsPanel(BottomPanel):

    def __init__(self, parent):
        BottomPanel.__init__(self, parent)
        Skin.set_background_color(self)
        self.layout = fsui.HorizontalLayout()

        def get_min_width():
            return 0
        # def get_min_height():
        #     return Constants.SCREEN_SIZE[1] + 2 * BORDER
        self.layout.get_min_width = get_min_width
        # self.layout.get_min_height = get_min_height
        self.layout.padding_left = BORDER // 2
        self.layout.padding_right = BORDER // 2
        self.layout.padding_top = BORDER + 2
        self.layout.padding_bottom = Skin.get_bottom_margin()

        self.default_image = fsui.Image("fs_uae_launcher:res/screenshot.png")
        # self.default_image.resize(Constants.SCREEN_SIZE)
        self.screenshot_overlay = fsui.Image(
            "fs_uae_launcher:res/screenshot_overlay.png")

        self.images = [self.default_image for _ in range(6)]
        self.image_paths = ["" for _ in range(6)]
        self.requests = [None for _ in range(6)]

        self.x_offset = 0

        if Settings.get("config_base"):
            # self.load_images(Settings.get("config_base"))
            self.load_images()
        Settings.add_listener(self)

    def on_destroy(self):
        Settings.remove_listener(self)

    def set_min_screenshots(self, count):
        # w = SCREEN_SIZE[0] * count + BORDER * 2 + (BORDER + 1) * (count - 1)
        w = Constants.SCREEN_SIZE[0] * count + BORDER * 2 + BORDER * (count - 1)
        self.set_min_width(w)

    def load_images(self):
        # t1 = time.time()
        handler = GamePaths.current()
        for i in range(6):
            path = handler.get_screenshot_path(i)
            if path == self.image_paths[i]:
                continue
            self.image_paths[i] = path

            loader = ImageLoader.get()

            def on_load(request):
                # print("on_load, request.image =", request.image)
                if request.path != self.image_paths[request.args["index"]]:
                    return
                if request.image:
                    self.images[request.args["index"]] = request.image
                else:
                    self.images[request.args["index"]] = self.default_image
                self.refresh()

            self.requests[i] = loader.load_image(
                path, size=Constants.SCREEN_SIZE, on_load=on_load, index=i)
            self.images[i] = self.default_image
            self.refresh()

            # image = handler.load_screenshot_preview(i)
            # if image:
            #     self.images[i] = image
            #     self.refresh()
            # else:
            #     self.images[i] = self.default_image
            #     self.refresh()

        # t2 = time.time()
        # print(t2 - t1)

    def on_setting(self, key, value):
        unused(value)
        if key == "config_name":
            self.x_offset = 0
            self.load_images()
        if key == "parent_uuid":
            self.x_offset = 0
            self.load_images()

    def on_left_down(self):
        # print("on_left_down")
        width = 22 + Constants.SCREEN_SIZE[0] * 6 + 22 * 5
        if self.x_offset == 0:
            self.x_offset = self.size[0] - width
        else:
            self.x_offset = 0
        self.refresh()

    def on_paint(self):
        dc = self.create_dc()
        self.draw_background(dc)
        size = self.size

        y = 2 + 20
        x = 10 + self.x_offset
        for i in range(6):
            if x >= size[0] - 12:
                break
            image = self.images[i]
            # dc.draw_image(image, x + 1, y + 1, Constants.SCREEN_SIZE[0],
            #         Constants.SCREEN_SIZE[1])
            dc.draw_image(image, x + 1, y + 1)
            dc.draw_image(self.screenshot_overlay, x - 10, y - 10)
            x = x + Constants.SCREEN_SIZE[0] + 22



import json
import time
from gzip import GzipFile
from io import StringIO
from urllib.request import HTTPBasicAuthHandler, build_opener, Request
from fsgs.res import gettext
from .client import OGDClient


class SynchronizerBase(object):

    def __init__(self, context, on_status=None, stop_check=None):
        self.context = context
        self.on_status = on_status
        self._stop_check = stop_check

    @staticmethod
    def bytes_to_int(m):
        return m[0] << 24 | m[1] << 16 | m[2] << 8 | m[3]

    def stop_check(self):
        if self._stop_check:
            return self._stop_check()

    def set_status(self, title, status=""):
        if self.on_status:
            self.on_status((title, status))

    @staticmethod
    def get_server():
        return OGDClient.get_server()

    def get_opener(self):
        server = self.get_server()
        auth_handler = HTTPBasicAuthHandler()
        auth_handler.add_password(
            realm="Open Amiga Game Database",
            uri="http://{0}".format(server),
            user=self.context.username,
            passwd=self.context.password)
        opener = build_opener(auth_handler)
        return opener

    def fetch_json_attempt(self, url):
        data = self.fetch_data_attempt(url, accept_gzip_encoding=True)
        return json.loads(data.decode("UTF-8"))

    def fetch_json(self, url):
        for i in range(20):
            try:
                return self.fetch_json_attempt(url)
            except Exception as e:
                print(e)
                sleep_time = 2.0 + i * 0.3
                # FIXME: change second {0} to {1}
                self.set_status(
                    gettext("Download failed (attempt {0}) - retrying in {0} "
                            "seconds").format(i + 1, int(sleep_time)))
                time.sleep(sleep_time)
                self.set_status(
                    gettext("Retrying last operation (attempt {0})").format(
                        i + 1))
        return self.fetch_json_attempt(url)

    def fetch_data_attempt(self, url, accept_gzip_encoding=False):
        if not url.startswith("http"):
            url = "http://{0}{1}".format(self.get_server(), url)
        opener = self.get_opener()
        print(url)
        request = Request(url)
        if accept_gzip_encoding:
            request.add_header("Accept-Encoding", "gzip")
        response = opener.open(request)
        # print(response.headers)
        data = response.read()

        try:
            getheader = response.headers.getheader
        except AttributeError:
            getheader = response.getheader
        content_encoding = getheader("content-encoding", "").lower()
        if content_encoding == "gzip":
            fake_stream = StringIO(data)
            data = GzipFile(fileobj=fake_stream).read()
        return data

    def fetch_data(self, url):
        for i in range(10):
            try:
                return self.fetch_data_attempt(url)
            except Exception as e:
                print(e)
                sleep_time = 2.0 + i * 0.3
                # FIXME: change second {0} to {1}
                self.set_status(
                    gettext("Download failed (attempt {0}) - retrying in {0} "
                            "seconds").format(i + 1, int(sleep_time)))
                time.sleep(sleep_time)
                self.set_status(
                    gettext("Retrying last operation (attempt {0})").format(
                        i + 1))
        return self.fetch_data_attempt(url)



from fsgs.ogd.client import OGDClient
import fsui
from fsbc.Application import app
from fs_uae_workspace.shell import SimpleApplication, shell_open
from fs_uae_launcher.res import gettext


class LoginWindow(fsui.Dialog):

    def __init__(self):
        super().__init__(None, gettext("Log In to Your OAGD.net Account"))
        self.set_icon(fsui.Icon("password", "pkg:fs_uae_workspace"))

        self.layout = fsui.VerticalLayout()
        self.layout.set_padding(20, 20, 20, 20)

        heading_layout = fsui.HorizontalLayout()
        self.layout.add(heading_layout)
        heading_layout.add(fsui.ImageView(
            self, fsui.Image("fs_uae_workspace:res/48/password.png")))
        heading_layout.add_spacer(20)
        heading_layout_2 = fsui.VerticalLayout()
        heading_layout.add(
            heading_layout_2, expand=True, fill=False, valign=0.5)
        heading_layout_2.add(fsui.HeadingLabel(
            self, gettext("Log In to Your OAGD.net Account")))
        heading_layout_2.add_spacer(2)
        heading_layout_2.add(fsui.Label(
            self, gettext("Logging in will enable the online game database "
                          "and more")))

        self.username_field = fsui.TextField(
            self, app.settings["database_email"])
        self.password_field = fsui.PasswordField(self)
        if self.username_field.get_text():
            self.password_field.focus()

        self.layout.add_spacer(20)
        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)
        label = fsui.Label(self, gettext("E-mail:"))
        label.set_min_width(100)
        hori_layout.add(label)
        hori_layout.add_spacer(20)
        # self.username_field.select_all()
        self.username_field.changed.connect(self.on_text_field_changed)
        self.username_field.activated.connect(self.on_username_activated)
        hori_layout.add(self.username_field, expand=True)

        self.layout.add_spacer(10)
        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)
        label = fsui.Label(self, gettext("Password:"))
        label.set_min_width(100)
        hori_layout.add(label)
        hori_layout.add_spacer(20)
        self.password_field.changed.connect(self.on_text_field_changed)
        self.password_field.activated.connect(self.on_password_activated)
        hori_layout.add(self.password_field, expand=True)

        self.layout.add_spacer(20)
        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)
        label = fsui.Label(self, gettext("Don't have an account already?"))
        hori_layout.add(label)
        hori_layout.add_spacer(20)
        label = fsui.URLLabel(
            self, gettext("Create an account now"),
            "http://oagd.net/user/register?referrer=fs-uae-launcher")
        hori_layout.add(label, expand=True)

        self.layout.add_spacer(6)
        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)
        label = fsui.Label(self,
                           gettext("Forgot your password?"))
        hori_layout.add(label)
        hori_layout.add_spacer(20)
        self.reset_label = fsui.URLLabel(
            self, gettext("Reset password via e-mail"),
            "http://oagd.net/user/reset?referrer=fs-uae-launcher")
        hori_layout.add(self.reset_label, expand=True)

        self.layout.add_spacer(20)
        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)
        self.created_label = fsui.Label(self, "")
        hori_layout.add(self.created_label, expand=True)
        hori_layout.add_spacer(20)

        self.login_button = fsui.Button(self, gettext("Log In"))
        self.login_button.disable()
        self.login_button.activated.connect(self.on_login_activated)
        hori_layout.add(self.login_button)

        self.close_button = fsui.Button(self, gettext("Close"))
        self.close_button.activated.connect(self.on_close_activated)
        hori_layout.add(self.close_button, margin_left=10)

        self.set_size(self.layout.get_min_size())
        self.center_on_parent()

    def __del__(self):
        print("LoginWindow.__del__")

    def on_close_activated(self):
        self.close()

    def on_text_field_changed(self):
        email = self.username_field.get_text().strip()
        password = self.password_field.get_text().strip()
        self.login_button.enable(bool(email) and bool(password))
        self.reset_label.set_url(
            "http://oagd.net/user/reset?referrer=fs-uae-launcher"
            "&email={0}".format(email))

    def on_username_activated(self):
        if self.username_field.get_text().strip():
            self.password_field.focus()

    def on_password_activated(self):
        self.on_login_activated()

    def on_login_activated(self):
        email = self.username_field.get_text().strip()
        password = self.password_field.get_text().strip()
        if not email or not password:
            return
        self.username_field.disable()
        self.password_field.disable()
        self.login_button.disable()

        task = OGDClient().login_task(email, password)
        # task.progressed.connect(self.progress)
        task.succeeded.connect(self.on_success)
        # task.failed.connect(fsui.error_function(gettext("Login Failed")))
        task.failed.connect(self.on_failure)
        task.start()

    def on_success(self):
        center = self.get_window_center()

        fsui.call_after(start_refresh_task, center)
        # shell_open("Workspace:Tools/Refresh", center=self.get_window_center())
        self.close()

    def on_failure(self, message):
        fsui.show_error(message, parent=self.get_window())
        self.username_field.enable()
        self.password_field.enable()
        self.login_button.enable()
        self.password_field.select_all()
        self.password_field.focus()

    # def on_progress(self):
    #    print("on_progress")

    # def on_success(self):
    #     self.close()


def start_refresh_task(center):
    shell_open("Workspace:Tools/Refresh", center=center)


application = SimpleApplication(LoginWindow)



'''OpenGL extension VERSION.GL_1_3

This module customises the behaviour of the 
OpenGL.raw.GL.VERSION.GL_1_3 to provide a more 
Python-friendly API

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/VERSION/GL_1_3.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
from OpenGL.GL import glget
import ctypes
from OpenGL.raw.GL import _types
from OpenGL.raw.GL.VERSION.GL_1_3 import *
from OpenGL.raw.GL.VERSION.GL_1_3 import _EXTENSION_NAME

def glInitGl13VERSION():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )

### END AUTOGENERATED SECTION
GL_SRC0_ALPHA = GL_SOURCE0_ALPHA # alias
GL_SRC0_RGB = GL_SOURCE0_RGB # alias
GL_SRC1_ALPHA = GL_SOURCE1_ALPHA # alias
GL_SRC1_RGB = GL_SOURCE1_RGB # alias
GL_SRC2_ALPHA = GL_SOURCE2_ALPHA # alias
GL_SRC2_RGB = GL_SOURCE2_RGB # alias


for typ,arrayType in (
    ('d',arrays.GLdoubleArray),
    ('f',arrays.GLfloatArray),
    ('i',arrays.GLintArray),
    ('s',arrays.GLshortArray),
):
    for size in (1,2,3,4):
        name = 'glMultiTexCoord%(size)s%(typ)sv'%globals()
        globals()[name] = arrays.setInputArraySizeType(
            globals()[name],
            size,
            arrayType,
            'v',
        )
        try:
            del size,name
        except NameError as err:
            pass
    try:
        del typ,arrayType
    except NameError as err:
        pass

for typ,arrayType in (
    ('d',arrays.GLdoubleArray),
    ('f',arrays.GLfloatArray),
):
    for function in ('glLoadTransposeMatrix','glMultTransposeMatrix'):
        name = '%s%s'%(function,typ)
        globals()[name] = arrays.setInputArraySizeType(
            globals()[name],
            16,
            arrayType,
            'm',
        )
        try:
            del function,name
        except NameError as err:
            pass
    try:
        del typ,arrayType
    except NameError as err:
        pass

from OpenGL import wrapper
from OpenGL.raw.GL.VERSION import GL_1_3 as _simple
from OpenGL.GL import images, glget

for dimensions in (1,2,3):
    for function in ('glCompressedTexImage%sD','glCompressedTexSubImage%sD'):
        name = function%(dimensions,)
        globals()[ name ] = images.compressedImageFunction(
            getattr( _simple, name )
        )
        try:
            del name, function
        except NameError as err:
            pass
    try:
        del dimensions
    except NameError as err:
        pass

if _simple.glGetCompressedTexImage:
    def glGetCompressedTexImage( target, level, img=None ):
        """Retrieve a compressed texture image"""
        if img is None:
            length = glget.glGetTexLevelParameteriv(
                target, 0,
                _simple.GL_TEXTURE_COMPRESSED_IMAGE_SIZE_ARB,
            )
            img = arrays.ArrayDataType.zeros( (length,), constants.GL_UNSIGNED_BYTE )
        return _simple.glGetCompressedTexImage(target, 0, img);



import os
from fsgs.ChecksumTool import ChecksumTool
import fsui as fsui
from ...Config import Config
from ...I18N import gettext
from fsgs.FSGSDirectories import FSGSDirectories
from ..IconButton import IconButton
from ..LauncherFilePicker import LauncherFilePicker


class KickstartGroup(fsui.Group):

    def __init__(self, parent):
        fsui.Group.__init__(self, parent)
        self.layout = fsui.VerticalLayout()

        heading_label = fsui.HeadingLabel(self, gettext("Kickstart ROM"))
        self.layout.add(heading_label, margin=10)
        self.layout.add_spacer(0)

        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)

        # label = fsui.Label(self, _("Kickstart ROM") + ":")
        # hori_layout.add(label, margin_left=10, margin_right=10)

        kickstart_types = [gettext("Default"), gettext("Custom"),
                           gettext("Internal")]
        self.kickstart_type_choice = fsui.Choice(self, kickstart_types)
        hori_layout.add(self.kickstart_type_choice, margin=10)

        self.text_field = fsui.TextField(self, "", read_only=True)
        hori_layout.add(self.text_field, expand=True, margin=10)

        self.browse_button = IconButton(self, "browse_file_16.png")
        self.browse_button.set_tooltip(gettext("Browse for File"))
        self.browse_button.activated.connect(self.on_browse_button)
        hori_layout.add(self.browse_button, margin=10)

        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)

        label = fsui.Label(self, gettext("Extended ROM:"))
        hori_layout.add(label, margin_left=10, margin_right=10)
        # self.layout.add_spacer(0)

        kickstart_types = [gettext("Default"), gettext("Custom")]
        self.ext_rom_type_choice = fsui.Choice(self, kickstart_types)
        hori_layout.add(self.ext_rom_type_choice, margin_right=10)

        self.ext_text_field = fsui.TextField(self, "", read_only=True)
        hori_layout.add(self.ext_text_field, expand=True, margin_right=10)

        self.ext_browse_button = IconButton(self, "browse_file_16.png")
        self.ext_browse_button.set_tooltip(gettext("Browse for File"))
        self.ext_browse_button.activated.connect(self.on_ext_browse_button)
        hori_layout.add(self.ext_browse_button, margin_right=10)

        self.initialize_from_config()
        self.set_config_handlers()

    def initialize_from_config(self):
        self.on_config("kickstart_file", Config.get("kickstart_file"))
        self.on_config("kickstart_ext_file", Config.get("kickstart_ext_file"))

    def set_config_handlers(self):
        self.kickstart_type_choice.on_change = self.on_kickstart_type_change
        self.ext_rom_type_choice.on_change = self.on_ext_rom_type_change
        Config.add_listener(self)

    def on_destroy(self):
        print("on_destroy")
        Config.remove_listener(self)

    def on_kickstart_type_change(self):
        index = self.kickstart_type_choice.get_index()
        if index == 0:
            if Config.get("kickstart_file") == "":
                return
            Config.set("kickstart_file", "")
        elif index == 2:
            if Config.get("kickstart_file") == "internal":
                return
            Config.set("kickstart_file", "internal")
        else:
            Config.set("kickstart_file", Config.get("x_kickstart_file"))
        Config.update_kickstart()

    def on_ext_rom_type_change(self):
        index = self.ext_rom_type_choice.get_index()
        if index == 0:
            if Config.get("kickstart_ext_file") == "":
                return
            Config.set("kickstart_ext_file", "")
        else:
            Config.set("kickstart_ext_file",
                       Config.get("x_kickstart_ext_file"))
        Config.update_kickstart()

    def on_browse_button(self, extended=False):
        default_dir = FSGSDirectories.get_kickstarts_dir()
        if extended:
            title = gettext("Choose Extended ROM")
            key = "kickstart_ext_file"
        else:
            title = gettext("Choose Kickstart ROM")
            key = "kickstart_file"
        dialog = LauncherFilePicker(self.get_window(), title, "rom",
                                    Config.get(key))
        if not dialog.show_modal():
            return
        path = dialog.get_path()

        checksum_tool = ChecksumTool(self.get_window())
        sha1 = checksum_tool.checksum_rom(path)

        dir, file = os.path.split(path)
        if extended:
            self.ext_text_field.set_text(file)
        else:
            self.text_field.set_text(file)
        if os.path.normcase(os.path.normpath(dir)) == \
                os.path.normcase(os.path.normpath(default_dir)):
            path = file

        if extended:
            Config.set_multiple([
                ("kickstart_ext_file", path),
                ("x_kickstart_ext_file", path),
                ("x_kickstart_ext_file_sha1", sha1)])
        else:
            Config.set_multiple([
                ("kickstart_file", path),
                ("x_kickstart_file", path),
                ("x_kickstart_file_sha1", sha1)])

    def on_ext_browse_button(self):
        return self.on_browse_button(extended=True)

    def on_config(self, key, value):
        if key == "kickstart_file":
            if value == "internal":
                self.text_field.set_text("")
                self.kickstart_type_choice.set_index(2)
            elif value:
                dir, file = os.path.split(value)
                self.text_field.set_text(file)
                self.kickstart_type_choice.set_index(1)
            else:
                self.text_field.set_text("")
                self.kickstart_type_choice.set_index(0)
        elif key == "kickstart_ext_file":
            if value:
                dir, file = os.path.split(value)
                self.ext_text_field.set_text(file)
                self.ext_rom_type_choice.set_index(1)
            else:
                self.ext_text_field.set_text("")
                self.ext_rom_type_choice.set_index(0)



from .eventhelper import EventHelper
from .manager import DeviceManager


class EventListener:

    def __init__(self):
        self.helper = EventHelper()
        self.helper.init()
        self.device_names = {}
        self.manager = DeviceManager(None)

    def get_next_event(self):
        event = self.helper.get_next_event()
        if event is None:
            return None

        if event["type"] in ["joy-device-added", "mouse-device-added",
                             "keyboard-device-added"]:
            count = self.device_names.get(event["name"], 0) + 1
            self.device_names[event["name"]] = count
            if count > 1:
                event["id"] = "{0} #{1}".format(event["name"], count)
            else:
                event["id"] = event["name"]

            if event["type"] == "joy-device-added":
                self.manager.add_joystick_device(event)
            elif event["type"] == "mouse-device-added":
                self.manager.add_mouse_device(event)
            elif event["type"] == "keyboard-device-added":
                self.manager.add_keyboard_device(event)

        return event



from fs_uae_launcher.ui.behaviors.configbehavior import ConfigBehavior
from fs_uae_launcher.ui.settings.overridewarning import OverrideWarning
import fsui as fsui
from ...I18N import gettext
from ..settings.FullscreenToggleButton import FullscreenToggleButton


class LaunchGroup(fsui.Group):

    def __init__(self, parent, add_label=False):
        fsui.Group.__init__(self, parent)
        self.layout = fsui.VerticalLayout()
        if add_label:
            label = fsui.Label(self, gettext("Launch FS-UAE"))
            self.layout.add(label)

        self.hori_layout = fsui.HorizontalLayout()
        self.layout.add(self.hori_layout, fill=True, expand=True)

        self.hori_layout.add_spacer(0, expand=True)

        self.hori_layout.add(OverrideWarning(self, "fullscreen"),
                             margin_right=10)
        self.hori_layout.add(FullscreenToggleButton(self), fill=True)

        self.start_button = fsui.Button(parent, gettext("Start"))
        self.start_button.activated.connect(self.on_start_button)
        self.hori_layout.add(self.start_button, fill=True, margin_left=10)

        ConfigBehavior(self, ["fullscreen"])

    def on_start_button(self):
        from ...FSUAELauncher import FSUAELauncher
        FSUAELauncher.start_game()

    def on_fullscreen_config(self, value):
        self.hori_layout.update()

    def set_min_height(self, height):
        pass



import fsui as fsui
from fsui.extra.iconheader import NewIconHeader
from ...I18N import gettext
from .OptionUI import OptionUI


class NetplaySettingsPage(fsui.Panel):

    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        self.layout = fsui.VerticalLayout()
        # self.layout.set_padding(20, 20, 20, 20)

        self.icon_header = NewIconHeader(
            self, fsui.Icon("netplay-settings", "pkg:fs_uae_workspace"),
            gettext("Net Play Settings"),
            "")
        self.layout.add(self.icon_header, fill=True, margin_bottom=20)

        def add_option(name):
            self.layout.add(OptionUI.create_group(self, name), fill=True,
                            margin_top=10, margin_bottom=10)

        add_option("irc_nick")
        add_option("netplay_tag")
        add_option("irc_server")



import os
import platform
import traceback
from configparser import ConfigParser, NoSectionError
from operator import itemgetter, attrgetter, methodcaller
from fsbc.system import windows, linux, macosx
from fsgs.FSGSDirectories import FSGSDirectories

class Plugin:

    def __init__(self, name, version, path, cp):
        self.name = name
        self.version = version
        self.path = path
        self.provides = {}
        self.load_provides(cp)

    def load_provides(self, cp):
        print("loading provides for", self.path, self.platform())
        try:
            for key, value in cp.items(self.platform()):
                self.provides[key] = value
        except NoSectionError:
            pass

    @staticmethod
    def platform():
        if windows:
            os_name = "windows"
        elif macosx:
            os_name = "macosx"
        elif linux:
            if os.environ.get("STEAM_RUNTIME", ""):
                os_name = "steamos"
            else:
                os_name = "linux"
        else:
            raise Exception("unknown plugin os")
        if platform.machine().lower() in ["x86_64", "x86-64", "amd64", "i386", "i486", "i586", "i686"]:
            if platform.architecture()[0] == "64bit":
                arch = "x86-64"
            else:
                arch = "x86"
        else:
            raise Exception("unknown plugin arch")
        return os_name + "_" + arch


class PluginResource:

    def __init__(self, plugin, name):
        self.plugin = plugin
        self.name = name

    @property
    def path(self):
        p = os.path.join(self.plugin.path, self.plugin.provides[self.name])
        return p


class PluginExecutable:

    def __init__(self, plugin, path):
        self.plugin = plugin
        self.path = path


class PluginManager:

    __instance = None

    def __init__(self):
        self.provides = {}
        self.plugins = []
        self.load_plugins()
        for plugin in self.plugins:
            for key, value in plugin.provides.items():
                self.provides[key] = plugin

    def load_plugins(self):
        plugins_dir = FSGSDirectories.get_plugins_dir()
        for name in os.listdir(plugins_dir):
            plugin_dir = os.path.join(plugins_dir, name)
            try:
                plugin = self.load_plugin(plugin_dir)
            except Exception:
                traceback.print_exc()
                continue
            if plugin is None:
                continue
            self.plugins.append(plugin)
        self.plugins.sort(key=attrgetter("name"))

    def load_plugin(self, plugin_dir):
        plugin_ini = os.path.join(plugin_dir, "plugin.ini")
        name = os.path.basename(plugin_dir).split("_")[0]
        if not os.path.exists(plugin_ini):
            return None
        cp = ConfigParser()
        with open(plugin_ini, "r", encoding="UTF-8") as f:
            cp.read_file(f)
        version = cp.get("plugin", "version")
        plugin = Plugin(name, version, os.path.join(plugin_dir, version), cp)
        return plugin

    @classmethod
    def instance(cls):
        if cls.__instance is None:
            cls.__instance = cls()
        return cls.__instance

    def find_resource(self, name):
        plugin = self.provides[name]
        return PluginResource(plugin, name)

    def find_executable(self, name):
        print("PluginManager.find_executable", name)
        if windows:
            name = "{0}.exe".format(name)
        elif macosx:
            name = "{0}.app/Contents/MacOS/{0}".format(name)

        for plugin in self.plugins:
            path = os.path.join(plugin.path, Plugin.platform(), name)
            print("check", path)
            if os.path.exists(path):
                return PluginExecutable(plugin, path)



from .mess import MESSRunner


class AmstradCPCRunner(MESSRunner):

    CONTROLLER = {
        "type": "controller",
        "description": "Controller",
        "mapping_name": "amstradcpc",
    }

    PORTS = [
        {
            "description": "1st Controller",
            "types": [CONTROLLER]
        }, {
            "description": "2nd Controller",
            "types": [CONTROLLER]
        },
    ]

    def mess_configure(self):
        self.mess_configure_floppies(["floppydisk1", "floppydisk2"])
        if self.config["command"]:
            pass
            self.inject_fake_input_string(60, "{0}\n".format(
                self.config["command"]))
            # self.inject_fake_input_string(60, "aaaaaaaaaaaaaaaaa")

    def mess_full_keyboard(self):
        return True

    def mess_input_mapping(self, port):
        return {
            "UP": "P#_JOYSTICK_UP",
            "DOWN": "P#_JOYSTICK_DOWN",
            "LEFT": "P#_JOYSTICK_LEFT",
            "RIGHT": "P#_JOYSTICK_RIGHT",
            "1": "P#_BUTTON1",
            "2": "P#_BUTTON2",
        }

    def get_game_refresh_rate(self):
        return 50.080128

    def mess_romset(self):
        return "cpc464", CPC464_ROMS


CPC464_ROMS = {
    "56d39c463da60968d93e58b4ba0e675829412a20": "cpc464.rom",
    "39102c8e9cb55fcc0b9b62098780ed4a3cb6a4bb": "cpcados.rom",
}



from fsgs.platform import PlatformHandler
from fsgs.mednafen.game_boy_advance import GameBoyAdvanceRunner
from .loader import SimpleLoader


class GameBoyAdvancePlatformHandler(PlatformHandler):

    PLATFORM_NAME = "Game Boy Advance"

    def __init__(self):
        PlatformHandler.__init__(self)

    def get_loader(self, fsgs):
        return GameBoyAdvanceLoader(fsgs)

    def get_runner(self, fsgs):
        return GameBoyAdvanceRunner(fsgs)


class GameBoyAdvanceLoader(SimpleLoader):
    pass



import io
import os
import hashlib
import traceback
from configparser import ConfigParser, NoSectionError
from fsgs.context import fsgs
from fsgs.ChecksumTool import ChecksumTool
from fsgs.amiga.Amiga import Amiga
from fsgs.amiga.ValueConfigLoader import ValueConfigLoader
from fsgs.FSGSDirectories import FSGSDirectories
from .Settings import Settings
from fsbc.signal import Signal
from .Util import expand_path


# the order of the following keys is significant (for some keys).
# multiple options should be set in this order since some options will
# implicitly change other options. examples:
# - amiga model can change kickstart and joystick modes
# - file options should be set before corresponding sha1 options
from fsgs.platform import PlatformHandler

cfg = [
    ("amiga_model",           "",     "checksum", "sync"),
    ("ntsc_mode",             "",         "checksum", "sync"),
    ("accuracy",              "",         "checksum", "sync"),
    ("chip_memory",           "",         "checksum", "sync"),
    ("slow_memory",           "",         "checksum", "sync"),
    ("fast_memory",           "",         "checksum", "sync"),
    ("zorro_iii_memory",      "",         "checksum", "sync"),
    ("bsdsocket_library",     "",         "checksum", "sync"),
    ("uaegfx_card",           "",         "checksum", "sync"),
    ("joystick_port_0",       ""),
    ("joystick_port_0_mode",  "",         "checksum", "sync"),
    ("joystick_port_0_autofire",  "",     "checksum", "sync"),
    ("joystick_port_1",       ""),
    ("joystick_port_1_mode",  "",         "checksum", "sync"),
    ("joystick_port_1_autofire",  "",     "checksum", "sync"),
    ("joystick_port_2",       ""),
    ("joystick_port_2_mode",  "",         "checksum", "sync"),
    ("joystick_port_2_autofire",  "",     "checksum", "sync"),
    ("joystick_port_3",       ""),
    ("joystick_port_3_mode",  "",         "checksum", "sync"),
    ("joystick_port_3_autofire",  "",     "checksum", "sync"),

    ("floppy_drive_count",    "",         "checksum", "sync"),
    ("cdrom_drive_count",     "",         "checksum", "sync"),

    # this is not an Amiga device, so no need to checksum / sync
    ("joystick_port_4_mode",  "",         "custom"),

    ("kickstart_file",        ""),
    ("x_kickstart_file",      "",                             "nosave"),
    ("x_kickstart_file_sha1", "",         "checksum", "sync", "nosave"),
    ("kickstart_ext_file",    ""),
    ("x_kickstart_ext_file",  "",                             "nosave"),
    ("x_kickstart_ext_file_sha1", "",     "checksum", "sync", "nosave"),

    ("x_whdload_args",        "",         "checksum", "sync"),
    ("x_whdload_version",     "",     "checksum", "sync"),
    ("floppy_drive_count",    "",         "checksum", "sync", "custom"),
    ("floppy_drive_speed",    "",         "checksum", "sync", "custom"),
    ("cdrom_drive_count",     "",         "checksum", "sync", "custom"),
    ("dongle_type",           "",         "checksum", "sync", "custom"),

    ("__netplay_game",        "",         "checksum", "sync"),
    ("__netplay_password",    "",         "checksum", "sync"),
    ("__netplay_players",     "",         "checksum", "sync"),
    ("__netplay_port",        "",                     "sync"),
    ("__netplay_addresses",   "",         "checksum", "sync"),
    ("__netplay_host",        ""),
    ("__netplay_ready",       ""),
    ("__netplay_state_dir_name",  "",     "checksum", "sync"),
    ("__version",             "FIXME"),
    ("__error",               ""),
    ("x_game_uuid",           ""),
    ("x_game_xml_path",       ""),
    ("title",                 "",                             "custom"),
    ("sub_title",             "",                             "custom"),
    ("viewport",              "",                             "custom"),

    ("year",                  ""),
    ("developer",             ""),
    ("publisher",             ""),
    ("languages",             ""),
    ("players",               ""),
    ("protection",            ""),
    ("hol_url",               ""),
    ("wikipedia_url",         ""),
    ("database_url",          ""),
    ("lemon_url",             ""),
    ("mobygames_url",         ""),
    ("amigamemo_url",         ""),
    ("whdload_url",           ""),
    ("mobygames_url",         ""),
    ("longplay_url",          ""),
    ("__variant_rating",      ""),
    ("variant_rating",        ""),
    ("variant_uuid",          ""),
    
    ("download_file",         ""),
    ("download_page",         ""),
    ("download_terms",        ""),
    ("download_notice",       ""),

    ("x_missing_files",       ""),
    ("x_game_notice",         ""),
    ("x_variant_notice",      ""),
    ("x_variant_warning",      ""),
    ("x_variant_error",      ""),
    ("x_joy_emu_conflict",    ""),

    ("screen1_sha1",          ""),
    ("screen2_sha1",          ""),
    ("screen3_sha1",          ""),
    ("screen4_sha1",          ""),
    ("screen5_sha1",          ""),
    ("front_sha1",            ""),
    ("title_sha1",            ""),
]

for _i in range(Amiga.MAX_FLOPPY_DRIVES):
    cfg.append(("floppy_drive_{0}".format(_i), ""))
    cfg.append(("x_floppy_drive_{0}_sha1".format(_i),
                "", "checksum", "sync", "nosave"))
for _i in range(Amiga.MAX_FLOPPY_IMAGES):
    cfg.append(("floppy_image_{0}".format(_i), ""))
    cfg.append(("x_floppy_image_{0}_sha1".format(_i),
                "", "checksum", "sync", "nosave"))
for _i in range(Amiga.MAX_CDROM_DRIVES):
    cfg.append(("cdrom_drive_{0}".format(_i), ""))
    cfg.append(("x_cdrom_drive_{0}_sha1".format(_i),
                "", "checksum", "sync", "nosave"))
for _i in range(Amiga.MAX_CDROM_IMAGES):
    cfg.append(("cdrom_image_{0}".format(_i), ""))
    cfg.append(("x_cdrom_image_{0}_sha1".format(_i),
                "", "checksum", "sync", "nosave"))
for _i in range(Amiga.MAX_HARD_DRIVES):
    cfg.append(("hard_drive_{0}".format(_i), ""))
    cfg.append(("hard_drive_{0}_label".format(_i),
                "", "checksum", "sync", "custom"))
    cfg.append(("hard_drive_{0}_priority".format(_i),
                "", "checksum", "sync", "custom"))
    cfg.append(("x_hard_drive_{0}_sha1".format(_i),
                "", "checksum", "sync", "nosave"))


class Config(object):

    config_keys = [x[0] for x in cfg]

    default_config = {}
    for c in cfg:
        default_config[c[0]] = c[1]

    key_order = [x[0] for x in cfg]
    checksum_keys = [x[0] for x in cfg if "checksum" in x]
    sync_keys_list = [x[0] for x in cfg if "sync" in x]
    sync_keys_set = set(sync_keys_list)
    no_custom_config = [x[0] for x in cfg if "custom" not in x]

    no_custom_config.append("__changed")
    no_custom_config.append("__config_name")
    no_custom_config.append("__database")
    no_custom_config.append("__ready")
    no_custom_config.append("x_whdload_icon")
    no_custom_config.append("platform")

    dont_save_keys_set = set([x[0] for x in cfg if "nosave" in x])

    reset_values = {}
    for i in range(Amiga.MAX_FLOPPY_DRIVES):
        reset_values["floppy_drive_{0}".format(i)] = \
            ("x_floppy_drive_{0}_sha1".format(i), "")
    for i in range(Amiga.MAX_FLOPPY_IMAGES):
        reset_values["floppy_image_{0}".format(i)] = \
            ("x_floppy_image_{0}_sha1".format(i), "")
    for i in range(Amiga.MAX_CDROM_DRIVES):
        reset_values["cdrom_drive_{0}".format(i)] = \
            ("x_cdrom_drive_{0}_sha1".format(i), "")
    for i in range(Amiga.MAX_CDROM_IMAGES):
        reset_values["cdrom_image_{0}".format(i)] = \
            ("x_cdrom_image_{0}_sha1".format(i), "")
    for i in range(Amiga.MAX_HARD_DRIVES):
        reset_values["hard_drive_{0}".format(i)] = \
            ("x_hard_drive_{0}_sha1".format(i), "")
    reset_values["x_kickstart_file"] = ("x_kickstart_file_sha1", "")
    reset_values["x_kickstart_ext_file"] = ("x_kickstart_ext_file_sha1", "")

    # config = default_config.copy()
    # config_listeners = []

    @classmethod
    def copy(cls):
        return fsgs.config.copy()

    @classmethod
    def get(cls, key, default=""):
        return fsgs.config.get(key, default)

    @classmethod
    def add_listener(cls, listener):
        # deprecated
        Signal("fsgs:config").connect(getattr(listener, "on_config"))

    @classmethod
    def remove_listener(cls, listener):
        # deprecated
        Signal("fsgs:config").disconnect(getattr(listener, "on_config"))

    @classmethod
    def set(cls, key, value):
        fsgs.config.set(key, value)

    @classmethod
    def set_multiple(cls, items):
        fsgs.config.set(items)

    @classmethod
    def update_from_config_dict(cls, config_dict):
        changes = []
        for key, value in config_dict.items():
            if key in fsgs.config.values:
                if fsgs.config.values[key] != value:
                    changes.append((key, value))
            else:
                changes.append((key, value))
        cls.set_multiple(changes)

    @classmethod
    def sync_items(cls):
        for key, value in fsgs.config.values.items():
            if key in cls.sync_keys_set:
                yield key, value

    @classmethod
    def checksum(cls):
        return cls.checksum_config(fsgs.config.copy())

    @classmethod
    def checksum_config(cls, config):
        s = hashlib.sha1()
        for key in cls.checksum_keys:
            value = config[key]
            s.update(str(value).encode("UTF-8"))
        return s.hexdigest()

    @classmethod
    def update_kickstart_in_config_dict(cls, config_dict):
        print("update_kickstart_in_config")
        model = config_dict.setdefault(
            "amiga_model", cls.default_config["amiga_model"])

        kickstart_file = config_dict.setdefault("kickstart_file", "")
        if kickstart_file:
            config_dict["x_kickstart_file"] = config_dict["kickstart_file"]
            if kickstart_file == "internal":
                config_dict["x_kickstart_file_sha1"] = Amiga.INTERNAL_ROM_SHA1
            else:
                # FIXME: set checksum
                pass
        else:
            checksums = Amiga.get_model_config(model)["kickstarts"]
            for checksum in checksums:
                path = fsgs.file.find_by_sha1(checksum)
                if path:
                    config_dict["x_kickstart_file"] = path
                    config_dict["x_kickstart_file_sha1"] = checksum
                    break
            else:
                print("WARNING: no suitable kickstart file found")
                config_dict["x_kickstart_file"] = ""
                config_dict["x_kickstart_file_sha1"] = Amiga.INTERNAL_ROM_SHA1

        if config_dict.setdefault("kickstart_ext_file", ""):
            config_dict["x_kickstart_ext_file"] = \
                config_dict["kickstart_ext_file"]
            # FIXME: set checksum
        else:
            checksums = Amiga.get_model_config(model)["ext_roms"]
            if len(checksums) == 0:
                config_dict["x_kickstart_ext_file"] = ""
                config_dict["x_kickstart_ext_file_sha1"] = ""
            else:
                for checksum in checksums:
                    path = fsgs.file.find_by_sha1(checksum)
                    if path:
                        config_dict["x_kickstart_ext_file"] = path
                        config_dict["x_kickstart_ext_file_sha1"] = checksum
                        break
                else:
                    # print("WARNING: no suitable kickstart ext file found")
                    config_dict["x_kickstart_ext_file"] = ""
                    config_dict["x_kickstart_ext_file_sha1"] = ""
                    # Warnings.set("hardware", "kickstart_ext",
                    #              "No suitable extended kickstart found")
                    # FIXME: set sha1 and name x_options also

    @classmethod
    def update_kickstart(cls):
        cls.set_kickstart_from_model()

    @classmethod
    def set_kickstart_from_model(cls):
        print("set_kickstart_from_model")
        config_dict = fsgs.config.values.copy()
        cls.update_kickstart_in_config_dict(config_dict)
        cls.update_from_config_dict(config_dict)

    @classmethod
    def load_default_config(cls):
        print("load_default_config")
        cls.load({})
        # FIXME: remove use of config_base
        Settings.set("config_base", "")
        Settings.set("config_name", "Unnamed Configuration")
        Settings.set("config_path", "")
        Settings.set("config_xml_path", "")

    @classmethod
    def load(cls, config):
        update_config = {}
        for key, value in cls.default_config.items():
            update_config[key] = value
        for key in list(fsgs.config.values.keys()):
            if key not in cls.default_config:
                # this is not a recognized key, so we remove it
                del fsgs.config.values[key]

        for key, value in config.items():
            # if this is a settings key, change settings instead
            if key in Settings.initialize_from_config:
                Settings.set(key, value)
            else:
                update_config[key] = value

        cls.update_kickstart_in_config_dict(update_config)
        cls.fix_loaded_config(update_config)
        # print("about to set", update_config)
        cls.set_multiple(update_config.items())
        # Settings.set("config_changed", "0")
        cls.set("__changed", "0")

        # cls.update_kickstart()

    @classmethod
    def fix_joystick_ports(cls, config):
        # from .Settings import Settings

        print("---", config["joystick_port_0"])
        print("---", config["joystick_port_1"])

        from .DeviceManager import DeviceManager
        available = DeviceManager.get_joystick_names()
        available.extend(["none", "mouse", "keyboard"])
        available_lower = [x.lower() for x in available]

        device_ids = DeviceManager.get_joystick_ids()
        device_ids.extend(["none", "mouse", "keyboard"])

        # remove devices from available list if specified and found
        try:
            index = available_lower.index(config["joystick_port_1"].lower())
        except ValueError:
            pass
        else:
            del available[index]
            del available_lower[index]
        try:
            index = available_lower.index(config["joystick_port_0"].lower())
        except ValueError:
            pass
        else:
            del available[index]
            del available_lower[index]

        # if config in
        # print("--------------------------------------------")
        if config["joystick_port_1_mode"] in ["joystick", "cd32 gamepad"]:
            if not config["joystick_port_1"]:
                want = Settings.get("primary_joystick").lower()
                # print("want", want)
                try:
                    index = available_lower.index(want)
                except ValueError:
                    index = -1
                print("available", available_lower)
                print("want", repr(want), "index", index)
                if index == -1:
                    index = len(available) - 1
                if index >= 0:
                    config["joystick_port_1"] = device_ids[index]
                    del available[index]
                    del available_lower[index]
                    del device_ids[index]
        elif config["joystick_port_1_mode"] in ["mouse"]:
            if not config["joystick_port_1"]:
                config["joystick_port_1"] = "mouse"
        elif config["joystick_port_1_mode"] in ["nothing"]:
            if not config["joystick_port_1"]:
                config["joystick_port_1"] = "none"

        if config["joystick_port_0_mode"] in ["joystick", "cd32 gamepad"]:
            if not config["joystick_port_0"]:
                want = Settings.get("secondary_joystick").lower()
                try:
                    index = available_lower.index(want)
                except ValueError:
                    index = -1
                # print("want", want, "index", index)
                if index == -1:
                    index = len(available) - 1
                if index >= 0:
                    config["joystick_port_0"] = device_ids[index]
                    del available[index]
                    del available_lower[index]
                    del device_ids[index]
        elif config["joystick_port_0_mode"] in ["mouse"]:
            if not config["joystick_port_0"]:
                config["joystick_port_0"] = "mouse"
        elif config["joystick_port_0_mode"] in ["nothing"]:
            if not config["joystick_port_0"]:
                config["joystick_port_0"] = "none"

    @classmethod
    def fix_loaded_config(cls, config):
        # cls.fix_joystick_ports(config)

        # FIXME: parent
        checksum_tool = ChecksumTool(None)

        def fix_file_checksum(sha1_key, key, base_dir, is_rom=False):
            path = config.get(key, "")
            # hack to synchronize URLs
            # print(repr(path))
            if path.startswith("http://") or path.startswith("https://"):
                sha1 = path
                config[sha1_key] = sha1
                return
            path = expand_path(path)
            sha1 = config.get(sha1_key, "")
            if not path:
                return
            if sha1:
                # assuming sha1 is correct
                return
            if not os.path.exists(path):
                print(repr(path), "does not exist")
                path = os.path.join(base_dir, path)
                if not os.path.exists(path):
                    print(repr(path), "does not exist")
                    return
            if os.path.isdir(path):
                # could set a fake checksum here or something, to indicate
                # that it isn't supposed to be set..
                return
            print("checksumming", repr(path))
            size = os.path.getsize(path)
            if size > 64 * 1024 * 1024:
                # not checksumming large files right now
                print("not checksumming large file")
                return

            if is_rom:
                sha1 = checksum_tool.checksum_rom(path)
            else:
                sha1 = checksum_tool.checksum(path)
            config[sha1_key] = sha1

        for i in range(Amiga.MAX_FLOPPY_DRIVES):
            fix_file_checksum(
                "x_floppy_drive_{0}_sha1".format(i),
                "floppy_drive_{0}".format(i),
                FSGSDirectories.get_floppies_dir())
        for i in range(Amiga.MAX_FLOPPY_IMAGES):
            fix_file_checksum(
                "x_floppy_image_{0}_sha1".format(i),
                "floppy_image_{0}".format(i),
                FSGSDirectories.get_floppies_dir())
        for i in range(Amiga.MAX_CDROM_DRIVES):
            fix_file_checksum(
                "x_cdrom_drive_{0}_sha1".format(i),
                "cdrom_drive_{0}".format(i),
                FSGSDirectories.get_cdroms_dir())
        for i in range(Amiga.MAX_CDROM_IMAGES):
            fix_file_checksum(
                "x_cdrom_image_{0}_sha1".format(i),
                "cdrom_image_{0}".format(i),
                FSGSDirectories.get_cdroms_dir())
        for i in range(Amiga.MAX_HARD_DRIVES):
            fix_file_checksum(
                "x_hard_drive_{0}_sha1".format(i),
                "hard_drive_{0}".format(i),
                FSGSDirectories.get_hard_drives_dir())

        # FIXME: need to handle checksums for Cloanto here
        fix_file_checksum(
            "x_kickstart_file_sha1", "x_kickstart_file",
            FSGSDirectories.get_kickstarts_dir(), is_rom=True)
        fix_file_checksum(
            "x_kickstart_ext_file_sha1", "x_kickstart_ext_file",
            FSGSDirectories.get_kickstarts_dir(), is_rom=True)

    @classmethod
    def load_file(cls, path):
        try:
            cls._load_file(path, "")
        except Exception:
            # FIXME: errors should be logged / displayed
            cls.load_default_config()
            traceback.print_exc()

    @classmethod
    def load_data(cls, data):
        print("Config.load_data")
        try:
            cls._load_file("", data)
        except Exception:
            # FIXME: errors should be logged / displayed
            cls.load_default_config()
            traceback.print_exc()

    @classmethod
    def create_fs_name(cls, name):
        name = name.replace(':', ' - ')
        name = name.replace('*', '-')
        name = name.replace('?', '')
        name = name.replace('/', '-')
        name = name.replace('\\', '-')
        name = name.replace('"', "'")
        for i in range(3):
            name = name.replace('  ', ' ')
        while name.endswith('.'):
            name = name[:-1]
        name = name.strip()
        return name

    @classmethod
    def _load_file(cls, path, data):
        if data:
            print("loading config from data")
        else:
            print("loading config from " + repr(path))
            if not os.path.exists(path):
                print("config file does not exist")
        if data:
            raise Exception("_load_file (data) not implemented")
        # if data:
        #     config_xml_path = ""
        #     loader = XMLConfigLoader()
        #     loader.load_data(data)
        #     config = loader.get_config()
        # elif path.endswith(".xml"):
        #     config_xml_path = path
        #     loader = XMLConfigLoader()
        #     loader.load_file(path)
        #     config = loader.get_config()
        else:
            config_xml_path = ""
            cp = ConfigParser(interpolation=None, strict=False)
            try:
                with io.open(path, "r", encoding="UTF-8") as f:
                    cp.readfp(f)
                # cp.read([path])
            except Exception as e:
                print(repr(e))
                return
            config = {}
            try:
                keys = cp.options("config")
            except NoSectionError:
                keys = []
            for key in keys:
                config[key] = cp.get("config", key)
            try:
                keys = cp.options("fs-uae")
            except NoSectionError:
                keys = []
            for key in keys:
                config[key] = cp.get("fs-uae", key)

        Settings.set("config_path", path)

        cls.load(config)

        config_name = config.get("__config_name", "")
        if config_name:
            config_name = cls.create_fs_name(config_name)
        else:
            config_name, ext = os.path.splitext(os.path.basename(path))

        if "(" in config_name:
            config_base = config_name.split("(")[0].strip()
        else:
            config_base = config_name
        # game = name

        # if not Config.get("title"):
        #     Config.set("title", config_base)

        Settings.set("config_base", config_base)
        Settings.set("config_name", config_name)
        Settings.set("config_xml_path", config_xml_path)
        # Settings.set("config_changed", "0")
        cls.set("__changed", "0")

    @classmethod
    def load_values(cls, values, uuid=""):
        print("loading config values", values)
        platform_id = values.get("platform", "").lower()

        if platform_id in ["amiga", "cdtv", "cd32"]:
            value_config_loader = ValueConfigLoader(uuid=uuid)
            value_config_loader.load_values(values)
            config = value_config_loader.get_config()
            cls.load(config)
            config_name = config.get("__config_name", "")

        else:
            print("Warning: Non-Amiga game loaded")
            platform_handler = PlatformHandler.create(platform_id)
            loader = platform_handler.get_loader(fsgs)
            fsgs.config.load(loader.load_values(values))
            config_name = "{0} ({1})".format(
                values.get("game_name"), values.get("platform_name"))

        Settings.set("config_path", "")

        # print("config is", config)
        # config["x_config_uuid"] = uuid

        if config_name:
            config_name = cls.create_fs_name(config_name)
        # else:
        #     config_name, ext = os.path.splitext(os.path.basename(path))

        if "(" in config_name:
            config_base = config_name.split("(")[0].strip()
        else:
            config_base = config_name
        # game = name

        # if not Config.get("title"):
        #     Config.set("title", config_base)

        Settings.set("config_base", config_base)
        Settings.set("config_name", config_name)
        Settings.set("config_xml_path", "")
        # Settings.set("config_changed", "0")
        cls.set("__changed", "0")



import os
from fsgs.FSGSDirectories import FSGSDirectories
from .GameNameUtil import GameNameUtil


class SaveStateHandler(object):

    def __init__(self, fsgs, name, platform, uuid=None):
        self.fsgs = fsgs
        self.uuid = uuid

        self.config_name = name
        if "(" in name:
            parts = name.split("(", 1)
            self.name, self.variant = parts
            self.name = self.name.strip()
            self.variant = self.variant.strip()
            if self.variant.endswith(")"):
                self.variant = self.variant[:-1]
            self.variant = self.variant.replace(") (", ", ")
            self.variant = self.variant.replace(")(", ", ")
        else:
            self.name = name
            self.variant = ""
        self.platform = platform

    def get_name(self):
        return self.name

    def get_variant(self):
        return self.variant

    def _get_state_dir(self):
        config_name = self.config_name
        if not config_name:
            config_name = "Default"

        # use a temporary state dir, for now, to avoid problems with
        # floppy overlays etc interfering with net play
        if self.fsgs.netplay.enabled:
            # it is possible to manually specify the state dir
            config_name = self.fsgs.config.get("__netplay_state_dir_name")
            if not config_name:
                # this is the default behavior, create a clean state
                # dir for the net play session
                netplay_game = self.fsgs.config.get("__netplay_game")
                if netplay_game:
                    config_name = "Net Play ({0})".format(netplay_game)

        # convert the config name to a name which can be represented on
        # the file system (really all/most filesystems).
        config_name = GameNameUtil.create_fs_name(config_name)

        letter = self.get_letter(config_name)
        if not letter:
            config_name = "Default"
            letter = self.get_letter(config_name)
        # we use an existing state dir in a "letter" dir if it exists
        # (legacy support).
        path = os.path.join(
            FSGSDirectories.get_save_states_dir(), letter, config_name)
        if os.path.exists(path):
            return path
        # if not, we use a direct subfolder of save states dir
        path = os.path.join(
            FSGSDirectories.get_save_states_dir(), config_name)
        return path

    def get_state_dir(self):
        state_dir = self._get_state_dir()
        if not os.path.exists(state_dir):
            os.makedirs(state_dir)
        return state_dir

    @staticmethod
    def get_letter(name):
        letter_name = name.upper()
        if letter_name.startswith("THE "):
            letter_name = letter_name[4:]
        if letter_name.startswith("A "):
            letter_name = letter_name[2:]
        for i in range(len(letter_name)):
            letter = letter_name[i]
            if letter in "01234567890":
                letter = "0"
                break
            if letter in "ABCDEFGHIJKLMNOPQRSTUVWXYZ":
                break
        else:
            return None
        return letter



from fsui.qt import QCheckBox, Signal
from .Widget import Widget


class CheckBox(QCheckBox, Widget):

    changed = Signal()

    def __init__(self, parent, text=""):
        # self._widget = QCheckBox(text, parent.get_container())
        QCheckBox.__init__(self, text, parent.get_container())
        # Widget.__init__(self, parent)
        self.init_widget(parent)
        # self._widget.stateChanged.connect(self.__state_changed)
        self.stateChanged.connect(self.__state_changed)

    def __state_changed(self):
        self.changed.emit()
        self.on_change()

    def is_checked(self):
        return self.isChecked()

    def check(self, checked=True):
        self.setChecked(checked)

    def on_change(self):
        pass


class HeadingCheckBox(CheckBox):

    def __init__(self, parent, text=""):
        CheckBox.__init__(self, parent, text)
        font = self.get_font()
        font.set_bold(True)
        self.set_font(font)



from fsgs.platform import PlatformHandler
from fsgs.mednafen.game_gear import GameGearRunner
from .loader import SimpleLoader


class GameGearPlatformHandler(PlatformHandler):

    PLATFORM_NAME = "Game Gear"

    def __init__(self):
        PlatformHandler.__init__(self)

    def get_loader(self, fsgs):
        return GameGearLoader(fsgs)

    def get_runner(self, fsgs):
        return GameGearRunner(fsgs)


class GameGearLoader(SimpleLoader):
    pass



'''OpenGL extension EXT.robustness

This module customises the behaviour of the 
OpenGL.raw.GLES2.EXT.robustness to provide a more 
Python-friendly API

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/EXT/robustness.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper

import ctypes
from OpenGL.raw.GLES2 import _types
from OpenGL.raw.GLES2.EXT.robustness import *
from OpenGL.raw.GLES2.EXT.robustness import _EXTENSION_NAME

def glInitRobustnessEXT():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )

### END AUTOGENERATED SECTION


import os
from fsbc.system import windows
from fsbc.task import current_task
from fsgs.input.mapper import InputMapper
from fsgs.runner import GameRunner


# noinspection PyAttributeOutsideInit
class MAMERunner(GameRunner):

    def init(self):
        self.mame_init()

    def mame_emulator_name(self):
        return "mame"

    def mame_init_input(self):
        """Override in inherited classes"""
        pass

    def mame_romset(self):
        """Override in inherited classes"""
        raise NotImplementedError()

    def init_input(self):
        self.mame_init_input()
        if windows:
            self.input_device_order = "DINPUT8"

    def prepare(self):
        self.configure()
        self.mame_prepare()

    def mame_init(self):
        # override in subclasses
        pass

    def mame_prepare(self):
        pass

    def configure(self):

        self.configure_romset()

        self.default_xml = ["""\ufeff<?xml version="1.0"?>
<!-- This file is autogenerated; comments and unknown tags will be stripped -->
<mameconfig version="10">
    <system name="default">\n"""]

        self.game_xml = ["""\ufeff<?xml version="1.0"?>
<!-- This file is autogenerated; comments and unknown tags will be stripped -->
<mameconfig version="10">
    <system name="{0}">\n""".format(self.romset)]

        self.add_arg("-skip_gameinfo")
        # state_dir = self.get_state_dir()
        state_dir = self.emulator_state_dir(self.mame_emulator_name())
        state_dir = state_dir + os.sep

        self.cwd_dir = self.create_temp_dir("mame-cwd")
        self.cfg_dir = self.create_temp_dir("mame-cfg")
        self.roms_dir = self.create_temp_dir("mame-roms")
        rom_path = self.roms_dir.path
        assert self.romset
        system_rom_path = os.path.join(rom_path, self.romset)
        os.makedirs(system_rom_path)

        for sha1, name in self.romset_files:
            file_uri = self.fsgs.file.find_by_sha1(sha1)
            current_task.set_progress("Preparing ROM {name}".format(name=name))
            input_stream = self.fsgs.file.open(file_uri)
            if input_stream is None:
                raise Exception("Cannot not find required ROM " + repr(name))
            path = os.path.join(system_rom_path, name)
            with open(path, "wb") as f:
                f.write(input_stream.read())

        # shutil.copy(self.get_game_file(),
        #             os.path.join(rom_path, self.romset + '.zip'))

        # MAME uses ; as path separator on all systems, apparently
        try:
            # self.args.extend(["-rompath", self.bios_dir()])
            rom_path = rom_path + ";" + self.mame_get_bios_dir()
        except:
            pass
        # rom_path = rom_path + ";" + os.path.dirname(self.get_game_file())

        # rom_path = rom_path + os.pathsep + os.path.dirname(
        #        self.get_game_file())
        self.add_arg("-rompath", rom_path)

        # copy initial nvram disk, if any, to nvram dir
        # p, e = os.path.splitext(self.get_game_file())
        # initram = p + ".nv"
        # if os.path.exists(initram):
        #    shutil.copy(initram, os.path.join(state_dir, self.romset + '.nv'))

        game_cfg_file = os.path.join(
            self.cfg_dir.path, "{romset}.cfg".format(romset=self.romset))
        self.add_arg("-cfg_directory", self.cfg_dir.path)
        self.add_arg("-nvram_directory", state_dir)
        # self.add_arg("-memcard_directory", state_dir)
        # self.add_arg("-hiscore_directory", state_dir)
        self.add_arg("-state_directory", state_dir)
        self.add_arg("-diff_directory", state_dir)

        self.add_arg("-snapshot_directory", self.screenshots_dir())
        self.add_arg("-snapname", "{0}-%i".format(self.screenshots_name()))

        # self.change_handler = GameChangeHandler(self.cfg_dir.path)
        # self.change_handler.init(
        #     os.path.join(self.get_state_dir(), "cfg"))

        self.configure_input()
        self.configure_video()

        self.game_xml.append('    </system>\n')
        self.game_xml.append('</mameconfig>\n')
        with open(game_cfg_file, 'wb') as f:
            f.write("".join(self.game_xml).encode("UTF-8"))
            print("")
            print("")
            print("".join(self.game_xml))
            print("")
            print("")

        self.default_xml.append('    </system>\n')
        self.default_xml.append('</mameconfig>\n')
        with open(os.path.join(self.cfg_dir.path, "default.cfg"), "wb") as f:
            f.write("".join(self.default_xml).encode("UTF-8"))
            print("")
            print("")
            print("".join(self.default_xml))
            print("")
            print("")

        if self.use_doubling():
            self.add_arg("-prescale", "2")

        if self.use_smoothing():
            self.add_arg("-filter")
        else:
            self.add_arg("-nofilter")

        # start mame/mess with UI keyboard keys enabled by default,
        # full keyboard can be activated with INS / Scroll-Lock
        # self.add_arg("-ui_active")

        self.mame_configure()

    def mame_configure(self):
        # override in subclasses
        pass

    def configure_romset(self):
        self.romset, self.romset_files = self.mame_romset()
        try:
            # convert from dict to list
            # noinspection PyUnresolvedReferences
            self.romset_files = list(self.romset_files.items())
        except AttributeError:
            # already a list
            pass

    def configure_input(self):
        # return
        # FIXME: ignoring input
        self.default_xml.append("        <input>\n")
        self.game_xml.append("        <input>\n")
        ports = {}
        # for i, input in enumerate(self.inputs):
        for i, port in enumerate(self.ports):
            input_mapping = self.mame_input_mapping(i)

            mapper = MAMEInputMapper(port, input_mapping)
            for key, value in mapper.items():
                print("---->", key, value)
                if isinstance(key, tuple):
                    key, type = key
                else:
                    type = 'standard'
                if 'type=' not in key:
                    key = 'type="{0}"'.format(key)
                key = key.replace("#", str(i + 1))
                # if '/' in key:
                #     key, tag = key.split('/')
                # else:
                #     tag = None
                # if ':' in key:
                #     key, type = key.split(':')
                # else:
                #     type = 'standard'
                if 'AD_STICK' in key: # and type == 'standard':
                    analog_axis = True
                else:
                    analog_axis = False
                if analog_axis and 'AXIS' in value:
                    value = value[:value.index('AXIS') + 4]
                    # remove increment / decrement type, set type
                    # to standard since this is an analog axis
                    type = 'standard'
                ports.setdefault(key, {}).setdefault(type, set()).add(value)
        for key, port in ports.items():
            # key, tag = key
            if 'tag=' in key:
                xml = self.game_xml
                # xml.append(
                #         '            <port tag="{tag}" '
                #         'type="{key}" mask="1" default="0">'
                #         '\n'.format(tag=tag, key=key))
                # xml.append(
                #         '            <port {key}>\n'.format(key=key))
            else:
                xml = self.default_xml
                # xml.append(
                #         '            <port type="{key}">\n'.format(key=key))
                # xml.append(
                #         '            <port {key}>\n'.format(key=key))
            xml.append('            <port {key}>\n'.format(key=key))
            for type, values in port.items():
                xml.append('                <newseq type="{type}">\n'.format(
                    type=type))
                xml.append('                    ')
                for i, value in enumerate(values):
                    if i > 0:
                        xml.append(' OR ')
                    xml.append(value)
                xml.append('\n                </newseq>\n')
            xml.append('            </port>\n')
        self.default_xml.append("        </input>\n")
        self.game_xml.append("        </input>\n")

    def configure_video(self):
        if windows and False:
            self.add_arg("-video", "d3d")
        else:
            self.add_arg("-video", "opengl")

        if self.use_fullscreen():
            self.add_arg("-nowindow")
        else:
            self.add_arg("-window")

        if self.use_stretching():
            self.add_arg("-nokeepaspect")
        else:
            self.add_arg("-keepaspect")

        self.game_xml.append('        <video>\n')
        self.game_xml.append('            <screen index="0" ')
        ox, oy, sx, sy = self.mame_offset_and_scale()
        self.game_xml.append('hstretch="{0:0.6f}" '.format(sx))
        self.game_xml.append('vstretch="{0:0.6f}" '.format(sy))
        self.game_xml.append('hoffset="{0:0.6f}" '.format(ox))
        self.game_xml.append('voffset="{0:0.6f}" '.format(oy))
        self.game_xml.append('/>\n')
        self.game_xml.append('        </video>\n')

        # effect = 'none'
        # filter_mapping = {
        #     'auto': 'aperture1x2rb',
        #     'rgb': 'aperture1x2rb',
        # }
        # for filter in self.context.config.filters:
        #     try:
        #         effect = filter_mapping[filter]
        #     except KeyError:
        #         continue
        #     break
        # self.args.extend(['-effect', effect])

        video_args = []
        if self.configure_vsync():
            video_override = "mame_video_options_vsync"
            video_args.append("-waitvsync")
            video_args.append("-syncrefresh")
            if windows and False:
                # no-throttle seems to enable turbo mode (no vsync throttling)
                # when used on Windows
                pass
            else:
                video_args.append("-nothrottle")
                pass
            # if self.get_game_refresh_rate():
            #     # should always be true since vsync was enabled...
            #     self.args.extend(
            #         ["-override_fps", str(self.get_game_refresh_rate())])
            if windows and False:
                video_args.append('-notriplebuffer')
        else:
            video_override = "mame_video_options"
            if windows and False:
                video_args.append('-triplebuffer')

        # print("mame_video_options_vsync", self.config[video_override])
        # raise Exception("gnit")

        if self.config[video_override]:
            for arg in self.config[video_override].split(" "):
                arg = arg.strip()
                if arg:
                    self.add_arg(arg)
        else:
            self.add_arg(*video_args)

    def mame_offset_and_scale(self):
        ox, oy, sx, sy = 0.0, 0.0, 1.0, 1.0
        viewport = self.config["viewport"]
        if viewport:
            src, dst = viewport.split("=")
            src = src.strip()
            dst = dst.strip()
            src_x, src_y, src_w, src_h = [int(v) for v in src.split(" ")]
            dst_x, dst_y, dst_w, dst_h = [int(v) for v in dst.split(" ")]
            sx = src_w / dst_w
            sy = src_h / dst_h
            ox = (((src_x + src_w / 2) - (dst_x + dst_w / 2)) / dst_w)
            oy = ((src_y + src_h / 2) - (dst_y + dst_h / 2)) / dst_h
        return ox, oy, sx, sy

    def mame_get_bios_dir(self):
        for dir in self.resources_dirs():
            bios_dir = os.path.join(dir, self.context.game.platform)
            if os.path.exists(bios_dir):
                return bios_dir
        raise Exception("Could not find bios dir")

    def run(self):
        self.add_arg(self.romset)
        # return self.start_emulator("fs-mame/mame")
        return self.start_emulator_from_plugin_resource("fs-mame")

    def finish(self):
        # if os.path.exists(os.path.join(self.cfg_dir.path, "default.cfg")):
        #     os.unlink(os.path.join(self.cfg_dir.path, "default.cfg"))
        # self.change_handler.update(
        #     os.path.join(self.get_state_dir(), "cfg"))
        pass


class MAMEInputMapper(InputMapper):

    def axis(self, axis, positive):
        axis_str = ["X", "Y", "Z", "RX", "RY", "RZ"][axis]
        if axis == 0:
            dir_str = "RIGHT" if positive else "LEFT"
        elif axis == 1:
            dir_str = "DOWN" if positive else "UP"
        else:
            # FIXME: Check this
            dir_str = "POS" if positive else "NEG"
        return "JOYCODE_%d_%sAXIS_%s_SWITCH" % (
            self.device.index + 1, axis_str, dir_str)

    def hat(self, hat, direction):
        if windows and False:
            dir_str = {
                'left': 'L',
                'right': 'R',
                'up': 'U',
                'down': 'D',
            }[direction]
            return "JOYCODE_{0}_HATSWITCH{1}".format(
                self.device.index + 1, dir_str)
        else:
            dir_str = {
                'left': 'LEFT',
                'right': 'RIGHT',
                'up': 'UP',
                'down': 'DOWN',
            }[direction]
            # in older MAME/MESS versions, hat 0 was specified here as 0,
            # but in recent versions, the first hat is called 1 in the config
            return "JOYCODE_{0}_HAT{1}{2}".format(
                self.device.index + 1, hat + 1, dir_str)

    def button(self, button):
        return "JOYCODE_{0}_BUTTON{1}".format(
            self.device.index + 1, button + 1)

    def key(self, key):
        return "KEYCODE_" + keycodes[key.sdl_name[5:].upper()]


from .keycodes import keycodes



import fsui as fsui
from ...I18N import gettext
from .MemoryWidget import MemoryWidget


class MemoryGroup(fsui.Group):

    def __init__(self, parent):
        fsui.Group.__init__(self, parent)
        self.layout = fsui.VerticalLayout()

        gettext("Memory")
        gettext("Memory Configuration")
        heading_label = fsui.HeadingLabel(
            self, gettext("Override Installed Memory"))
        self.layout.add(heading_label, margin=10)
        self.layout.add_spacer(0)

        # hori_layout = fsui.HorizontalLayout()
        # self.layout.add(hori_layout, fill=True)

        vert_layout = fsui.VerticalLayout()
        # hori_layout.add(vert_layout, fill=True, expand=-1)
        self.layout.add(vert_layout, fill=True)

        widget = MemoryWidget(
            self, gettext("Chip RAM"), "chip_memory",
            [256, 512, 1024, 1536, 2048, 4096, 8192])
        vert_layout.add(widget, fill=True, margin=10)

        widget = MemoryWidget(self, gettext("Slow RAM"), "slow_memory",
                              [0, 512, 1024, 1536, 1792])
        vert_layout.add(widget, fill=True, margin=10)

        # vert_layout = fsui.VerticalLayout()
        # hori_layout.add(vert_layout, fill=True, expand=-1)

        widget = MemoryWidget(
            self, gettext("Zorro II Fast RAM"), "fast_memory",
            [0, 64, 128, 256, 512, 1024, 2048, 4096, 8192])
        vert_layout.add(widget, fill=True, margin=10)

        widget = MemoryWidget(
            self, gettext("Zorro III Fast RAM"), "zorro_iii_memory",
            [0, 1024, 2048, 4096, 8192, 16384, 32768, 65536, 131072,
             262144, 393216, 524288, 786432, 1048576])
        vert_layout.add(widget, fill=True, margin=10)

        widget = MemoryWidget(
            self, gettext("Motherboard Fast RAM"), "motherboard_ram",
            [0, 1024, 2048, 4096, 8192, 16384, 32768, 65536])
        vert_layout.add(widget, fill=True, margin=10)



from fsgs.platform import PlatformHandler
from fsgs.mednafen.master_system import MasterSystemRunner
from .loader import SimpleLoader


class MasterSystemPlatformHandler(PlatformHandler):

    PLATFORM_NAME = "Master System"

    def __init__(self):
        PlatformHandler.__init__(self)

    def get_loader(self, fsgs):
        return MasterSystemLoader(fsgs)

    def get_runner(self, fsgs):
        return MasterSystemRunner(fsgs)


class MasterSystemLoader(SimpleLoader):
    pass



import fsui as fsui
from ..Config import Config
from ..Settings import Settings
from .XMLControl import XMLControl
from .PickGameDialog import PickGameDialog


class EditorGroup(fsui.Group):

    def __init__(self, parent):
        fsui.Group.__init__(self, parent)
        self.layout = fsui.VerticalLayout()
        # self.layout.padding_left = 20
        # self.layout.padding_right = 20

        self.layout.add_spacer(20)

        self.label = fsui.HeadingLabel(self, "Edit Config")
        self.layout.add(self.label)

        self.layout.add_spacer(10)

        horlayout = fsui.HorizontalLayout()
        self.layout.add(horlayout, fill=True)

        self.game_button = fsui.Button(self, "Pick Game")
        self.game_button.activated.connect(self.on_game_button)
        horlayout.add(self.game_button, fill=True)

        horlayout.add_spacer(10)
        horlayout.add_spacer(0, expand=True)

        self.save_button = fsui.Button(self, "Save")
        self.save_button.activated.connect(self.on_save_button)
        horlayout.add(self.save_button, fill=True)

        self.layout.add_spacer(14)

        self.config_xml_ctrl = XMLControl(self)
        self.layout.add(self.config_xml_ctrl, fill=True, expand=True)

        self.layout.add_spacer(20)

        self.game_xml_ctrl = XMLControl(self)
        self.layout.add(self.game_xml_ctrl, fill=True, expand=True)

        self.layout.add_spacer(20)

        Settings.add_listener(self)
        Config.add_listener(self)

    def on_config(self, key, value):
        if key == "x_game_xml_path":
            self.game_xml_ctrl.set_path(value)

    def on_setting(self, key, value):
        if key == "config_xml_path":
            # self.load_config_xml(value)
            self.config_xml_ctrl.set_path(value)

    def on_save_button(self):
        self.config_xml_ctrl.save()
        self.game_xml_ctrl.save()

    def on_game_button(self):
        dialog = PickGameDialog(self.get_window())
        if dialog.show_modal():
            game_info = dialog.get_game_info()
        else:
            game_info = None
        dialog.destroy()
        if game_info is None:
            return
        print(game_info)
        self.config_xml_ctrl.connect_game(game_info)
